/*
 * linux.c
 * Copyright (C) 2003-2016, Gerdes Aktiengesellschaft. All rights reserved.
 *
 * Contact: Gerdes AG, Bergerwiesenstr. 9, 53340 Meckenheim, Germany, email: info@gerdes-ag.de
 */

#include "linux.h"

#if defined(_DRV_LINUX_)
	#ifndef __KERNEL__
	#define __KERNEL__
	#endif

	#ifndef MODULE
	#define MODULE
	#endif
#endif

#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0) && LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	#define __KERNEL24__
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	#define __KERNEL26__
#else
	#error "Invalid kernel version"
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
#include <linux/config.h>
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 33)
#include <linux/autoconf.h>
#else
#include <generated/autoconf.h>
#endif

#if defined(CONFIG_SMP)
	#define __SMP__
#endif

#include <linux/module.h>
#include <linux/sched.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,38)
#include <linux/smp_lock.h>
#endif
#include <linux/pci.h>
#include <linux/usb.h>
#include <linux/pm.h>
#include <linux/isapnp.h>
#include <linux/poll.h>
#include <linux/vmalloc.h>
#include <linux/init.h>
#include <asm/io.h>

#if defined(__KERNEL26__)
#include <linux/moduleparam.h>
#include <linux/statfs.h>
#elif defined(__KERNEL24__)
#include <linux/devfs_fs_kernel.h>
#endif

#if !defined(HAVE_COMPAT_IOCTL)
#define HAVE_COMPAT_IOCTL 0
#endif

#if !defined(HAVE_UNLOCKED_IOCTL)
#define HAVE_UNLOCKED_IOCTL 0
#endif

#if !defined(__SPIN_LOCK_UNLOCKED)
#define __SPIN_LOCK_UNLOCKED(x) SPIN_LOCK_UNLOCKED
#endif

extern void   LINUXAPI LinuxDbgPrint(WORD type, LPCSTR fmt, ...);
extern void   LINUXAPI DbgTrace(LPCSTR fmt, ...);
extern LPVOID LINUXAPI HeapAllocate(UINT size);
extern void   LINUXAPI HeapFree(LPVOID ptr);

#if 1 // defined(CONFIG_X86_64)
#define __MALLOC(x)       vmalloc(x)
#define __FREE(x)         vfree(x)
#else
#define __MALLOC(x)       kmalloc(x, GFP_KERNEL)
#define __FREE(x)         kfree(x)
#endif


///////////////////////////////////////////////////////////////////////////////
// Globale Daten
#define __LICENSE   "Proprietary"
#define __VERBOSE   0

int   heapFlags = 0;
int   verbose   = __VERBOSE;
char* tz        = "CET-1CEST,M3.5.0/2,M10.5.0/3";

#ifdef MODULE_LICENSE
MODULE_LICENSE(__LICENSE);
#else
const char __module_license[] __attribute__((section(".modinfo"))) = "license=" __LICENSE;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 25)
#ifdef CONFIG_USB
#undef CONFIG_USB
#endif
#ifdef CONFIG_USB_MODULE
#undef CONFIG_USB_MODULE
#endif
#endif

///////////////////////////////////////////////////////////////////////////////
// universelle Struktur für den Treiber-Kontext
static struct semaphore userSema;

typedef struct _PRIV_DATA {
	LPVOID			  pIrpHdl; // Zeiger auf das zugehörige IRPObject
	DWORD             minor;
	LPVOID            vptr[3];
	} PRIV_DATA;

static void InitPrivateData(PRIV_DATA* privData, LPVOID irpHdl)
{
	privData->pIrpHdl  = irpHdl;
	privData->minor    = 0;
}

static PRIV_DATA* AllocHWData(LPVOID irpHdl, PHARDW_DEV phw)
{
	PRIV_DATA* res;

	if ((res = (PRIV_DATA*)HeapAllocate(sizeof(PRIV_DATA))) != NULL) {
		InitPrivateData(res, irpHdl);
		if (phw != NULL) {
			res->vptr[0] = phw->devParms.ptr1;
			res->vptr[1] = phw->devParms.ptr2;
			res->vptr[2] = phw->devParms.ptr3;
			}
		}
	return(res);
}

static void FreeHWData(LPVOID ptr)
{
	HeapFree(ptr);
}

static PRIV_DATA* AllocFileData(LPVOID irpHdl, DWORD minor)
{
	PRIV_DATA* res;

	if ((res = (PRIV_DATA*)HeapAllocate(sizeof(PRIV_DATA))) != NULL) {
		InitPrivateData(res, irpHdl);
		res->minor = minor;
		}
	return(res);
}

static void FreeFileData(LPVOID ptr)
{
	HeapFree(ptr);
}

#ifndef __user
#define __user /* */
#endif

static Params* LAllocateIrp(LPVOID context, DWORD major, DWORD minor, const void __user* inBuf, DWORD inLen, DWORD outLen)
{
	Params*    irp;
	int        stat;
	PRIV_DATA* pPriv = (PRIV_DATA*)context;

	if ((irp = (Params*)__MALLOC(sizeof(Params) + (inLen > outLen ? inLen : outLen))) == NULL)
		return(NULL);

	if (inLen != 0 && inBuf != NULL && (stat = copy_from_user(&irp[1], inBuf, inLen)) != 0) {
		DbgTrace("copy_from_user failed (res=%d,dst=%p,src=%p,cnt=%d)", stat, &irp[1], inBuf, inLen);
		__FREE(irp);
		return(NULL);
		}

	irp->pIrpHdl     = pPriv->pIrpHdl;
	irp->pFileHdl    = context;
	irp->wakeFlag    = 0;
	irp->status      = 0;
	irp->ioCnt       = 0;
	irp->evt_maj     = major;
	irp->evt_min     = minor;
	irp->minor       = pPriv->minor;
	irp->ioBuffer    = (LPBYTE)&irp[1];
	irp->ioInLen     = inLen;
	irp->ioOutLen    = outLen;
	irp->Complete    = NULL;
	irp->Context     = NULL;
	irp->ref         = 1;
	LListInit(&irp->entry);
	return(irp);
}

static void LFreeIrp(Params* irp)
{
	if (InterlockedDecrement(&irp->ref) == 0) {
		LListRemoveEntry(&irp->entry);
		__FREE(irp);
		}
}

static void CompletionHelper(Params* irp)
{
	wake_up_interruptible((wait_queue_head_t*)irp->Context);
}

static int ExecuteIrp(struct file* fp, LPVOID context, DWORD major, DWORD minor, const void __user* inBuf, DWORD inLen, void __user* outBuf, DWORD outLen)
{
	Params*           irp;
	int               res, stat;
	wait_queue_head_t wait;
	PRIV_DATA*        pPriv = (PRIV_DATA*)context;

	res = -EFAULT;
	if (pPriv != NULL && (irp = LAllocateIrp(context, major, minor, inBuf, inLen, outLen)) != NULL) {
		res = -ERESTARTSYS;
		if (down_interruptible(&userSema) == 0) {
			init_waitqueue_head(&wait);
			irp->Complete = CompletionHelper;
			irp->Context  = &wait;
			(*IRPHandler)(irp);
			up(&userSema);

			wait_event_interruptible(wait, (irp->status != STATUS_PENDING));

			if (irp->status == STATUS_SUCCESS) {
				if (pPriv->pIrpHdl == NULL)
					pPriv->pIrpHdl = irp->pIrpHdl;
				if ((res = irp->ioCnt) != 0 && outLen != 0 && (stat = copy_to_user(outBuf, irp->ioBuffer, res)) != 0) {
					DbgTrace("copy_to_user failed (res=%d,dst=%p,src=%p,cnt=%d)", stat, outBuf, irp->ioBuffer, res);
					res = -EFAULT;
					}
				}
			else if (irp->status == STATUS_CANCELLED)
				res = -ERESTARTSYS;
			else if (irp->status == ERROR_INVALID_PARAMETER)
				res = -EINVAL;
			else
				res = -EIO;
			}

		LFreeIrp(irp);
		}

	if (res < 0)
		DbgTrace("ExecuteIrp failed with res=-%d (major=%D, minor=%D)", -res, major, minor);

	return(res);
}

void LINUXAPI LCompleteRequest(Params* irp)
{
	if (irp->Complete != NULL)
		(*irp->Complete)(irp);
}

///////////////////////////////////////////////////////////////////////////////
// Global stuff
extern LPCSTR __drvName;
STATUS (*IRPHandler)(PParams) LINUXAPI;

///////////////////////////////////////////////////////////////////////////////
// Exithandling
typedef struct {
	void   (*func)(LPVOID);
	LPVOID ctx;
	} EXITHANDLER;

static EXITHANDLER exitHandlers[32];
static int         exitCount = 0;
static int         inExit    = 0;
static int         drvActive = FALSE;

static void LINUXAPI atExit(void (*func)(LPVOID), LPVOID ctx)
{
	exitHandlers[exitCount].func = func;
	exitHandlers[exitCount].ctx  = ctx;
	exitCount++;
}

static void LINUXAPI doExit(void)
{
	inExit = 1;
	while (exitCount-- != 0)
		(*exitHandlers[exitCount].func)(exitHandlers[exitCount].ctx);
}

///////////////////////////////////////////////////////////////////////////////
// Debugging
void LINUXAPI LDbgPrint(LPCSTR hdr, LPCSTR txt)
{
	if (verbose)
   		printk("%s%s\n", hdr, txt);
}

///////////////////////////////////////////////////////////////////////////////
// Root Rechte
BOOL LINUXAPI LCheckForRoot(void)
{
	return(capable(CAP_SYS_ADMIN));
}

///////////////////////////////////////////////////////////////////////////////
// Time and Timer
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 7)
	#include <linux/jiffies.h>

	#define jif2ms(j) jiffies_to_msecs(j)
	#define ms2jif(m) msecs_to_jiffies(m)
#else
	#include <asm/div64.h>

	static inline u64 div64(u64 x, int y)
	{
		do_div(x, y);
		return(x);
	}

	static unsigned int jif2ms(const unsigned long jif)
	{
		return((unsigned int)div64(jif * (u64)1000, HZ));
	}

	static unsigned long ms2jif(const unsigned int ms)
	{
		return((unsigned long)div64(ms*(u64)HZ + 999, 1000));
	}
#endif

DWORD LINUXAPI LGetSystemTime(void)
{
#ifndef INITIAL_JIFFIES
#define INITIAL_JIFFIES 0
#endif

	return(jif2ms(jiffies - INITIAL_JIFFIES));
}

///////////////////////////////////////////////////////////////////////////////
// Sleep und Wakeup
void LINUXAPI LDrvSleep(DWORD msec)
{
	unsigned long t1, diff;

	for (t1 = jiffies, diff = ms2jif(msec); jiffies - t1 < diff; )
		;
}

///////////////////////////////////////////////////////////////////////////////
// IRQ - Interrupt und SMP Locking
#if !defined(CONFIG_SMP)
#define _CPU                 (1)
#else
#define _CPU                 smp_processor_id()
#endif

#define NO_CPU                 255
#define __ASSERT(x)            do { if (!(x)) *(LPDWORD)NULL = 0xDEADBEEF; } while (0)

static DWORD volatile imCpu  = NO_CPU;
static spinlock_t     imLock = __SPIN_LOCK_UNLOCKED(imLock);
static unsigned long  imFlags;

DWORD LINUXAPI LSMPLock(void)
{
	if (imCpu == NO_CPU || imCpu != (DWORD)_CPU) {
		spin_lock_irqsave(&imLock, imFlags);
		__ASSERT(imCpu == NO_CPU);
		}
	return(xchg(&imCpu, _CPU));
}

void LINUXAPI LSMPUnlock(DWORD cpu)
{
	__ASSERT(imCpu == (DWORD)_CPU);
	if ((imCpu = cpu) == NO_CPU)
		spin_unlock_irqrestore(&imLock, imFlags);
}

BOOL LINUXAPI LInInterrupt(void)
{
   return(in_interrupt() || imCpu != NO_CPU);
}

extern BOOL LINUXAPI _IRQEntry(LPVOID ISRObj);

#if defined(__KERNEL24__)

static void irqEntry(int irq, LPVOID isrObj, struct pt_regs* regs)
{
	_IRQEntry(isrObj);
}

#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 19)

static irqreturn_t irqEntry(int irq, LPVOID isrObj, struct pt_regs* regs)
{
	return(IRQ_RETVAL(_IRQEntry(isrObj)));
}

#else

static irqreturn_t irqEntry(int irq, LPVOID isrObj)
{
	return(IRQ_RETVAL(_IRQEntry(isrObj)));
}

#endif

BOOL LINUXAPI LIrqRequest(UINT irq, BOOL shared, LPCSTR descr, LPVOID ISRObj)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 22)
	return(!request_irq(irq, irqEntry, (shared ? SA_SHIRQ : 0), descr, ISRObj));
#else
	return(!request_irq(irq, irqEntry, (shared ? IRQF_SHARED : 0), descr, ISRObj));
#endif
}

void LINUXAPI LIrqFree(UINT irq, LPVOID ISRObj)
{
   free_irq(irq, ISRObj);
}

///////////////////////////////////////////////////////////////////////////////
// Interlocked... Funktionen
LONG LINUXAPI InterlockedIncrement(LPLONG val)
{
	LONG  res;
	DWORD cpu;

	cpu = LSMPLock();
	res = (*val += 1);
	LSMPUnlock(cpu);
	return(res);
}

LONG LINUXAPI InterlockedDecrement(LPLONG val)
{
	LONG  res;
	DWORD cpu;

	cpu = LSMPLock();
	res = (*val -= 1);
	LSMPUnlock(cpu);
	return(res);
}

///////////////////////////////////////////////////////////////////////////////
// Work Items
#if defined(__KERNEL24__)

#define __WORK_HANDLER(name)       name(LPVOID ctx)
#define __WORK_STRUCT(name, func)  struct tq_struct name = { routine: func }
#define __SCHEDULE_WORK(name)      schedule_task(name)

#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)

#define __WORK_HANDLER(name)       name(LPVOID ctx)
#define __WORK_STRUCT(name, func)  DECLARE_WORK(name, func, NULL)
#define __SCHEDULE_WORK(name)      schedule_work(name)

#else

#define __WORK_HANDLER(name)       name(struct work_struct* work)
#define __WORK_STRUCT(name, func)  DECLARE_WORK(name, func)
#define __SCHEDULE_WORK(name)      schedule_work(name)

#endif

///////////////////////////////////////////////////////////////////////////////
// Memory management
static LPVOID freeQueue = NULL;

LPVOID LINUXAPI LAllocateSystemMemory(DWORD size, BOOL lock)
{
	LPVOID res;

	if ((res = __MALLOC(size)) == NULL)
		DbgTrace("LAllocateSystemMemory(%d) failed", size);
	return(res);
}

static void __WORK_HANDLER(freeHandler)
{
	DWORD  cpu;
	LPVOID ptr;

	cpu = LSMPLock();
	while ((ptr = freeQueue) != NULL) {
		freeQueue = *(LPVOID*)ptr;
		LSMPUnlock(cpu);
		__FREE(ptr);
		cpu = LSMPLock();
		}
	LSMPUnlock(cpu);
}

void LINUXAPI LFreeSystemMemory(LPVOID ptr)
{
	static __WORK_STRUCT(freeTask, freeHandler);

	if (LInInterrupt()) {
		DWORD  cpu;

		cpu = LSMPLock();
		*(LPVOID*)ptr = freeQueue;
		freeQueue     = ptr;
		LSMPUnlock(cpu);
		__SCHEDULE_WORK(&freeTask);
		}
	else
		__FREE(ptr);
}

BOOL LINUXAPI LPageAllocate(LPVOID pDev, DWORD size, LPDWORD physAddr, LPVOID* linAddr)
{
	dma_addr_t phys;

	if (in_interrupt())
		DbgTrace("ERROR: Trying to allocate memory in interrupt");
	if ((*linAddr = pci_alloc_consistent((struct pci_dev*)pDev, size, &phys)) == NULL) {
		DbgTrace("ERROR: PageAllocate failed");
		return(FALSE);
		}
	*physAddr = (DWORD)phys;
	return(TRUE);
}

void LINUXAPI LPageFree(LPVOID pDev, DWORD size, DWORD physAddr, LPVOID linAddr)
{
	pci_free_consistent((struct pci_dev*)pDev, size, linAddr, (dma_addr_t)physAddr);
}

///////////////////////////////////////////////////////////////////////////////
// Systemzeit
//
// POSIX Timezone Format:
//
//    stdoffset[dst[offset],[start[/time],end[/time]]]
//
//    std und dst sind Zeichenfolgen aus beliebigen Zeichen ausser ':', '0'-'9', ',', '-', '+'
//    offset ist eine Zeitangabe als hh[:mm[:ss]], steht ein '-' davor, liegt die Zeitzone östlich von Greenwich
//    Wird dst angegeben, aber kein zweites offset, so wird stdoffset+01:00:00 als Default verwendet
//    start und end geben an, wann DST beginnt bzw. endet. Format ist:
//      Jn       Julian Day (1 <= n <= 365). Schaltjahre werden nicht berücksichtigt, d.h. der 01.03. ist immer Tag 60
//      n        Julian Day (0 <= n <= 365). Schaltjahre werden berücksichtigt
//      Mm.n.d   Der d-te Tag der n-ten Woche im m-ten Monat (1 <= m <= 12, 1 <= n <= 5, 0 <= d <= 6)
//               n=5 bezeichnet den letzten entsprechenden Tag im Monat (kann auch die 4-te Woche sein)
//               d=0 ist Sonntag
//    time gibt die Schaltzeit am jeweiligen Tag an (Default ist 02:00:00).
//
// Zur Auswertung wird die Zeit zunächst in der Standard-Zone berechnet (also UTC + stdoffset). Liegt dieser Wert zwischen
// start und end, wird dstoffset - stdoffset dazuaddiert.
//
// Am besten geht man daher anders vor: die eigene Timezone Struktur enthält/berechnet die Schaltzeiten als echte Unix Timestamps in
// UTC: man ermittelt zunächst das Kalenderjahr und berechnet daraus dann die beiden Timestamps. Dann vergleicht man die aktuelle Zeit
// mit diesen Werten und addiert das passende Offset (std/dst). Danach läuft alles wie gehabt.
//
// Auf der Südhalbkugel ist start > end; bei der Berechnung muss das berücksichtigt werden: in diesem Fall verwendet man für die Berechnung
// des start Timestamps das Folgejahr.
//
// Bleibt lediglich das Problem des Jahreswechsels; wenn UTC noch 2009 hat, kann es in der lokalen Zeitzone bereits 2010 sein. Geht man davon
// aus, das Sommerzeit immer addiert, sollte man zur Berechnung des aktuellen Jahres UTC+stdoffset verwenden.

typedef struct {
	BYTE  type;
	BYTE  m, n, d;
	LONG  julian;
	LONG  time;
	} DSTSPEC;

typedef struct {
	LONG    stdoffset;     // Seconds east of UTC
	LONG    dstdelta;      // Diff between DST and STD in seconds
	DSTSPEC dststart;      // Spec for DST start
	DSTSPEC dstend;        // Spec for DST end
	} TIMEZONE;

// Hack für Deutschland
// TODO: Timezone als Parameter übergeben
static TIMEZONE my_tz, dflt_tz = {
	1*60*60, // stdoffset
	1*60*60, // dstdelta
	{'M', 2, 28, 0, 0, 2*60*60},
	{'M', 9, 28, 0, 0, 3*60*60},
	};

static WORD   mdays1[] = {0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365};
static WORD   mdays2[] = {0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366};
static LPWORD leap[]   = {mdays1, mdays1, mdays2, mdays1, mdays1};

static LPCSTR _strchr(char c, LPCSTR str)
{
	while (*str && *str != c)
		str++;
	return(*str == c ? str : NULL);
}

static LPCSTR readDecimal(LPCSTR str, LPLONG val)
{
	*val = 0;
	while (*str >= '0' && *str <= '9')
		*val = *val * 10 + (long)(*str++ - '0');
	return(str);
}

static LPCSTR readOffset(LPCSTR str, LPLONG val)
{
	LONG tmp;

	str  = readDecimal(str, &tmp);
	*val = tmp*60*60;
	if (*str == ':') {
		str   = readDecimal(str+1, &tmp);
		*val += tmp*60;
		if (*str == ':') {
			str   = readDecimal(str+1, &tmp);
			*val += tmp;
			}
		}
	return(str);
}

static LPCSTR readSignedOffset(LPCSTR str, LPLONG val)
{
	if (*str == '-')
		str = readOffset(str+1, val);
	else if (*str == '+') {
		str  = readOffset(str+1, val);
		*val = - *val;
		}
	else {
		str  = readOffset(str, val);
		*val = - *val;
		}
	return(str);
}

static LPCSTR readDSTSpec(LPCSTR str, DSTSPEC* dst)
{
	LONG m=0, n=0, d=0;

	if (*str == 'j' || *str == 'J') {
		dst->type   = 'J';
		str         = readDecimal(str+1, &m);
		dst->julian = (m - 1) * 24*60*60;
		}
	else if (*str == 'm' || *str == 'M') {
		dst->type = 'M';
		str = readDecimal(str+1, &m);
		if (*str == '.') {
			str = readDecimal(str+1, &n);
			if (*str == '.')
				str = readDecimal(str+1, &d);
			}
		dst->m = (BYTE)m-1;
		dst->n = (BYTE)(n-1) * 7;
		dst->d = (BYTE)d;
		}
	else {
		dst->type   = ' ';
		str         = readDecimal(str+1, &m);
		dst->julian = m * 24*60*60;
		}

	dst->time = 2*60*60;
	if (*str == '/')
		str = readOffset(str+1, &dst->time);

	return(str);
}

static void parseTimezone(TIMEZONE* tz, LPCSTR desc)
{
	*tz = dflt_tz;

	// Name der Timezone skippen
	while (*desc && _strchr(*desc, "0123456789,:+-") == NULL)
		desc++;

	if (*desc && _strchr(*desc, "0123456789+-") != NULL) {
		// stdoffset lesen
		desc = readSignedOffset(desc, &tz->stdoffset);
		tz->dstdelta = 0;

		if (*desc && _strchr(*desc, "0123456789,:+-") == NULL) {
			// Name der Timezone skippen
			while (*desc && _strchr(*desc, "0123456789,:+-") == NULL)
				desc++;

			tz->dstdelta = tz->stdoffset + 60*60;
			if (*desc && _strchr(*desc, "0123456789+-") != NULL)
				desc = readSignedOffset(desc, &tz->dstdelta);
			if (*desc == ',')
				desc = readDSTSpec(desc+1, &tz->dststart);
			if (*desc == ',')
				desc = readDSTSpec(desc+1, &tz->dstend);
			tz->dstdelta    -= tz->stdoffset;
			tz->dstend.time -= tz->dstdelta;
			}
		}
}

static DWORD dstHelper(DWORD val, LPWORD mdays, DWORD wday, DSTSPEC* dst)
{
	DWORD days, wd;

	switch (dst->type) {
		case 'M': // dst enthält m/n/d (0<=m<12, n=[0,7,14,21,28], 0<=d<=6)
			days = mdays[dst->m];
			if ((wd = (wday + days) % 7) > dst->d)
				days += 7;
			days += dst->d - wd;
			days += dst->n;
			wd = mdays[dst->m+1];
			while (days >= wd)
				days -= 7;
			val += days*24*60*60;
			break;

		case 'J': // dst->julian enthält Wert von 0 bis 364, 29. Februar wird nicht gezählt
			val += dst->julian;
			if (mdays[2] == 60 && dst->julian >= 59*24*60*60)
				val += 24*60*60;
			break;

		default: // dst->julian enthält Wert von 0 bis 365, 29. Februar wird gezählt
			val += dst->julian;
			break;
		}
	return(val + dst->time);
}

static void getCurrentTime(DWORD val, short tim[6], short* isdst)
{
	val   += my_tz.stdoffset;
	*isdst = 0;

	// Sommerzeit-Berechnungen nur durchführen, wenn es eine Sommerzeit gibt!
	if (my_tz.dstdelta != 0) {
		DWORD tmp, idx, wday, dststart, dstend;

		// Jahresanfang berechnen. Dazu berechnen wir zunächst den Rest zu vollständigen Blöcken je 1461 Tage (4 Jahre inkl. Schalttag). Dieser Rest
		// teilt sich dann in zwei Jahre ohne Schalttag, eines mit Schalttag und ein weiteres ohne Schalttag ein.
		tmp = val % (1461*24*60*60);
		if (tmp < 365*24*60*60) {
			idx = 0;
			}
		else if (tmp < 730*24*60*60) {
			tmp -= 365*24*60*60;
			idx  = 1;
			}
		else if (tmp < 1096*24*60*60) {
			tmp -= 730*24*60*60;
			idx  = 2;
			}
		else {
			tmp -= 1096*24*60*60;
			idx  = 3;
			}
		tmp      = val - tmp;                                        // tmp enthält jetzt den Timestamp vom 01.01 00:00:00 des aktuellen Jahres
		wday     = (4 + tmp / (24*60*60)) % 7;                       // wday enthält den Wochentag des 01.01. des aktuellen Jahres (01.01.1970 war ein Donnerstag)
		dststart = dstHelper(tmp, leap[idx], wday, &my_tz.dststart); // DST Start berechnen
		dstend   = dstHelper(tmp, leap[idx], wday, &my_tz.dstend);   // DST Ende berechnen

		// Falls DST Start nach DST Ende liegt, sind wir auf der Südhalbkugel. Wir wiederholen die DST Start Berechnung für das Folgejahr
		if (dststart > dstend) {
			tmp  += leap[idx][12] * 24*60*60;
			wday  = (wday + leap[idx][12]) % 7;
			dststart = dstHelper(tmp, leap[idx+1], wday, &my_tz.dststart);
			}

		if (val >= dststart && val < dstend) {
			val   += my_tz.dstdelta;
			*isdst = 1;
			}
		}

	// Umrechnen auf Tage seit dem 1.3.1968. Das hat den Vorteil, daß der Schalttag nicht irgendwo mitten im Jahr liegt, sondern genau hinten dran.
	val += (365 + 365 - 28 - 31) * 24*60*60;

	// Uhrzeit ermitteln
	tim[5] = (short)(val % 60);
	val /= 60;

	tim[4] = (short)(val % 60);
	val /= 60;

	tim[3] = (short)(val % 24);
	val /= 24;

	// Jahr ausrechnen: Ein kompletter Schaltjahresblock hat 4*365+1 = 1461 Tage. Volle Jahre innerhalb eines solchen Blocks haben 365 Tage
	tim[0] = 1968 + (short)(4*(val / 1461));
	val %= 1461;

	if (val == 1460) {
		// Falls jetzt genau 1460 Tage übrig bleiben, handelt es sich um den 29.2. des nächsten Schaltjahres.
		tim[0] += 4;
		tim[1] = 2;
		tim[2] = 29;
		}
	else {
		// Andernfalls werden die voll enthaltenen 365 Tage Blöcke auf das Jahr addiert, der Rest in die Zahl der Tage seit dem 1.3. des Jahres.
		tim[0] += (short)(val / 365);
		val %= 365;

		// Relativ zum 1.3. teilt sich ein Jahr in 3 (naja) Blöcke zu je 153 Tagen (31 30 31 30 31) auf. Die Zahl der Tage innerhalb des Blocks wird ermittelt
		tim[1] = (short)(5*(val / 153));
		val %= 153;

		// Relativ zum Blockstart haben wir jetzt 3 Blöcke von je 61 Tagen (31 30). Die Zahl der Tage relativ zu diesem Subblock wird ermittelt
		tim[1] += (short)(2*(val / 61));
		val %= 61;

		// Jetzt verbleiben zwei Blöcke je 31 Tage. Die Zahl der Tage relativ zum Restblock wird ermittelt
		tim[1] += (short)(val / 31);
		val %= 31;

		// Der Rest ist der gesuchte Tag im Monat, aber statt 0-x brauchen wir 1-x+1
		tim[2] = (short)(val + 1);

		// Der Monat (0-11) muß auf (3-12, 1-2) umgerechnet werden. Falls wir in den Monaten 10 oder 11 (Januar/Februar) sind, muß das Jahr korrigiert werden.
		if ((tim[1] += 3) > 12) {
			tim[1] -= 12;
			tim[0] += 1;
			}
		}

}

// tim = year month day hour minute second
void LINUXAPI LGetCurrentTime(short tim[6])
{
	struct timeval tv;
	short          isdst;

	do_gettimeofday(&tv);
	getCurrentTime((DWORD)tv.tv_sec, tim, &isdst);
}

void LINUXAPI LGetCurrentTime2(unsigned int tim[2])
{
	struct timeval tv;

	do_gettimeofday(&tv);
	tim[0] = tv.tv_sec;
	tim[1] = tv.tv_usec;
}

///////////////////////////////////////////////////////////////////////////////
// Listenverwaltung
void LINUXAPI LListInit(PLIST_ENTRY list)
{
	list->prev = list;
	list->next = list;
}

void LINUXAPI LListAddTail(PLIST_ENTRY list, PLIST_ENTRY ent)
{
	ent->prev        = list->prev;
	ent->next        = list;
	list->prev->next = ent;
	list->prev       = ent;
}

void LINUXAPI LListRemoveEntry(PLIST_ENTRY ent)
{
	if (ent->next != ent) {
		ent->prev->next = ent->next;
		ent->next->prev = ent->prev;
		ent->prev       = ent;
		ent->next       = ent;
		}
}

PLIST_ENTRY LINUXAPI LListRemoveHead(PLIST_ENTRY list)
{
	PLIST_ENTRY res;

	if ((res = list->next) == list)
		return(NULL);
	LListRemoveEntry(res);
	return(res);
}

int LINUXAPI LListIsEmpty(PLIST_ENTRY list)
{
	return(list->next == list);
}

///////////////////////////////////////////////////////////////////////////////
// HW - USB diverses

#if (defined(CONFIG_USB) || defined(CONFIG_USB_MODULE))

   #if defined(__KERNEL24__)

   #define _usb_alloc_urb(x)                                usb_alloc_urb(x)
   #define _usb_submit_urb(x)                               usb_submit_urb(x)
   #define _usb_kill_urb(x)                                 usb_unlink_urb(x)
   #define _urb_suffix                                      /* */

	static void _usb_init_urb(struct urb* urb)
	{
		memset(urb, 0, sizeof(*urb));
		spin_lock_init(&urb->lock);
	}

   #else

   #define _usb_alloc_urb(x)                                usb_alloc_urb(x, GFP_ATOMIC)
   #define _usb_init_urb(x)                                 usb_init_urb(x)
   #define _usb_submit_urb(x)                               usb_submit_urb(x, GFP_ATOMIC)
   #define _usb_set_intfdata(p1, p2)                        usb_set_intfdata(p1, p2)
   #define _usb_get_intfdata(p1)                            usb_get_intfdata(p1)

   #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 8)
      #define _usb_kill_urb(x)    usb_unlink_urb(x)
      #define _urb_suffix         , int mem_flags
   #elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 17)
      #define _usb_kill_urb(x)    usb_kill_urb(x)
      #define _urb_suffix         , int mem_flags
   #else
      #define _usb_kill_urb(x)    usb_kill_urb(x)
      #define _urb_suffix         , gfp_t mem_flags
   #endif

   #endif

#define _usb_unlink_urb(x)                                   usb_unlink_urb(x)

#define _usb_sndisocpipe(p1, p2)                             usb_sndisocpipe(p1, p2)
#define _usb_sndbulkpipe(p1, p2)                             usb_sndbulkpipe(p1, p2)
#define _usb_sndintpipe(p1, p2)                              usb_sndintpipe(p1, p2)
#define _usb_sndctrlpipe(p1, p2)                             usb_sndctrlpipe(p1, p2)
#define _usb_rcvctrlpipe(p1, p2)                             usb_rcvctrlpipe(p1, p2)
#define _usb_get_current_frame_number(p1)                    usb_get_current_frame_number(p1)

#define _usb_register(x)                                     usb_register(x)
#define _usb_deregister(x)                                   usb_deregister(x)
#define _usb_free_urb(x)                                     usb_free_urb(x)
#define _usb_control_msg(p1, p2, p3, p4, p5, p6, p7, p8, p9) usb_control_msg(p1, p2, p3, p4, p5, p6, p7, p8, p9)
#define _usb_set_interface(p1, p2, p3)                       usb_set_interface(p1, p2, p3)
#define _interface_to_usbdev(p1)                             interface_to_usbdev(p1)

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 30)
#define _usb_reset_endpoint(d, e)                            do {} while(0)
#else
#define _usb_reset_endpoint(d, e)                            usb_reset_endpoint(d, e)
#endif

#else

#define _usb_alloc_urb(x)                                    (NULL)
#define _usb_init_urb(x)                                     do {} while(0)
#define _usb_submit_urb(x)                                   (-1)
#define _usb_kill_urb(x)                                     do {} while(0)
#define _usb_unlink_urb(x)                                   do {} while(0)
#define _usb_set_intfdata(p1, p2)                            do {} while(0)
#define _usb_get_intfdata(p1)                                (NULL)
#define _usb_reset_endpoint(d, e)                            do {} while(0)

#define _usb_sndisocpipe(p1, p2)                             (-1)
#define _usb_sndbulkpipe(p1, p2)                             (-1)
#define _usb_sndintpipe(p1, p2)                              (-1)
#define _usb_sndctrlpipe(p1, p2)                             (-1)
#define _usb_rcvctrlpipe(p1, p2)                             (-1)
#define _usb_get_current_frame_number(p1)                    (-1)

#define _usb_register(x)                                     (-1)
#define _usb_deregister(x)                                   do {} while(0)
#define _usb_free_urb(x)                                     do {} while(0)
#define _usb_control_msg(p1, p2, p3, p4, p5, p6, p7, p8, p9) (-1)
#define _usb_set_interface(p1, p2, p3)                       (-1)
#define _interface_to_usbdev(p1)                             (NULL)

#endif

// Interface zum Linux-Kernel
static LPVOID _UsbProbe(struct usb_device* dev, const struct usb_device_id* id)
{
   HARDW_DEV hw;
	LPVOID    ptr;

	memset(&hw, 0, sizeof(hw));
   hw.pModel = (PSUPHWDEV)id->driver_info;
   hw.pOsDev = dev;
   if ((ptr = DriverAddDevice(&hw)) == NULL) {
      DbgTrace("usb probe: DriverAddDevice failed");
      return(NULL);
   	}
   return(AllocHWData(ptr, &hw));
}

#if defined(__KERNEL24__)
static LPVOID OS_UsbProbe(struct usb_device* dev, unsigned int interface, const struct usb_device_id* id)
{
	LPVOID ptr;

	if (interface != 0) {
      DbgTrace("usb probe: additional interface");
      return(NULL);
   	}
	return(_UsbProbe(dev, id));
}
#else
static int OS_UsbProbe(struct usb_interface* iface, const struct usb_device_id* id)
{
	struct usb_device* dev = _interface_to_usbdev(iface);
	LPVOID             ptr;

	DbgTrace("UsbProbe, dev=%p, data=%p, id=%p", dev, dev_get_drvdata(&dev->dev), id);
	if (dev_get_drvdata(&dev->dev) == (void*)id) {
      DbgTrace("usb probe: additional interface");
      return(-1);
   	}
	else if ((ptr = _UsbProbe(dev, id)) == NULL)
      return(-1);

	dev_set_drvdata(&dev->dev, (void*)id);
	_usb_set_intfdata(iface, ptr);
	return(0);
}
#endif

#if defined(__KERNEL24__)
static void OS_UsbDisconnect(struct usb_device* dev, LPVOID drv_context)
{
#else
static void OS_UsbDisconnect(struct usb_interface* iface)
{
	LPVOID drv_context = _usb_get_intfdata(iface);
#endif

	if (inExit == 0)
		ExecuteIrp(NULL, drv_context, IRP_MJ_PNP, IRP_MN_REMOVE_DEVICE, NULL, 0, NULL, 0);
	FreeHWData(drv_context);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 12)

static int OS_UsbSuspend(struct usb_interface* iface, pm_message_t state)
{
	int res;

	if ((res = ExecuteIrp(NULL, _usb_get_intfdata(iface), IRP_MJ_POWER, IRP_ARG_POWER_DOWN, NULL, 0, NULL, 0)) < 0)
		res = -EBUSY;
	return(res);
}

static int OS_UsbResume(struct usb_interface* iface)
{
	int res;

	if ((res = ExecuteIrp(NULL, _usb_get_intfdata(iface), IRP_MJ_POWER, IRP_ARG_POWER_UP, NULL, 0, NULL, 0)) < 0)
		res = -EBUSY;
	return(res);
}

#endif

static struct usb_driver OS_usbDriver = {
	name:           NULL,
   probe:          OS_UsbProbe,
   disconnect:     OS_UsbDisconnect,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 12)
	suspend:        OS_UsbSuspend,
	resume:         OS_UsbResume,
#endif
   id_table:       NULL,
	};

static struct usb_device_id* OS_usbTable;
static LIST_ENTRY            pendingUrbs = {&pendingUrbs, &pendingUrbs};
static spinlock_t            pendingLock = __SPIN_LOCK_UNLOCKED(pendingLock);

#if !defined(USB_ISO_ASAP)
#define USB_ISO_ASAP          URB_ISO_ASAP
#endif

#if !defined(USB_ENDPOINT_HALT)
#define USB_ENDPOINT_HALT     0
#endif

static void LUsbUnregisterDrv(LPVOID ptr)
{
	LinuxDbgPrint(6, "Unregistering USB driver...");
	_usb_deregister(&OS_usbDriver);
	HeapFree(OS_usbTable);
}

static int LINUXAPI LUsbRegisterDrv(struct usb_driver* drv)
{
   int result;

	LinuxDbgPrint(6, "Registering USB driver...");
   if ((result = _usb_register(drv)) < 0) {
      LinuxDbgPrint(6, "   usb_register returned %D", result);
		HeapFree(OS_usbTable);
      return(result);
	   }
	atExit(LUsbUnregisterDrv, NULL);
   return(0);
}

static STATUS LUsbMapStatus(int status, int line)
{
	if (status == 0)
		return(STATUS_SUCCESS);

//	DbgTrace("LUsbMapStatus(%d, %d)", -status, line);
	switch (status) {
		case -EPIPE:
		case -ECONNRESET:
			return(STATUS_HALTED);

		case -ENOENT:
		case -EILSEQ:
		case -ESHUTDOWN:
		case -ENODEV:
			return(STATUS_NO_SUCH_DEVICE);

		default:
			return(STATUS_NO_SUCH_DEVICE);
		}
}

LPVOID LINUXAPI LUsbAllocBuffer(int size)
{
	return(kmalloc(size, GFP_ATOMIC|__GFP_DMA));
}

void LINUXAPI LUsbFreeBuffer(LPVOID ptr)
{
	kfree(ptr);
}

PURB LINUXAPI LUsbAllocUrb(int iso_packets)
{
	PURB urb;

	if ((urb = (PURB)kmalloc(sizeof(URB) + iso_packets*sizeof(struct _URB_ISO_PKT), GFP_ATOMIC)) == NULL)
		return(NULL);
	memset(urb, 0, sizeof(URB));
	LListInit(&urb->UrbHeader.Entry);
	urb->UrbHeader.Ref = 1;
	// ...
	return(urb);
}

void LINUXAPI LUsbFreeUrb(PURB urb)
{
	urb->UrbHeader.Complete = NULL;
	if (InterlockedDecrement(&urb->UrbHeader.Ref) == 0)
		kfree(urb);
}

static STATUS LUsbCompleteUrb(PURB urb, STATUS stat)
{
	urb->UrbHeader.Status = stat;
	if (urb->UrbHeader.Complete != NULL)
		(urb->UrbHeader.Complete)(urb);
	if (InterlockedDecrement(&urb->UrbHeader.Ref) == 0)
		kfree(urb);
	return(stat);
}

static int LUsbEndpointFromPipe(unsigned int pipe)
{
	int endp;

	endp = usb_pipeendpoint(pipe);
	if (usb_pipein(pipe))
		endp |= USB_DIR_IN;
	return(endp);
}

static struct urb* LUsbBuildControlRequest(struct usb_device* dev, unsigned int pipe, unsigned char requesttype, unsigned char request, unsigned short value, unsigned short index, void* buf, unsigned short len)
{
   struct usb_ctrlrequest* dr;
   struct urb*             urb;

	if ((urb = (struct urb*)kmalloc(sizeof(struct urb) + sizeof(struct usb_ctrlrequest), GFP_ATOMIC)) != NULL) {
		_usb_init_urb(urb);

		dr                          = (struct usb_ctrlrequest*)&urb[1];
		dr->bRequestType            = requesttype;
		dr->bRequest                = request;
		dr->wValue                  = cpu_to_le16(value);
		dr->wIndex                  = cpu_to_le16(index);
		dr->wLength                 = cpu_to_le16(len);

		urb->setup_packet           = (unsigned char*)dr;
		urb->dev                    = dev;
		urb->pipe                   = pipe;

		urb->transfer_buffer        = buf;
		urb->transfer_buffer_length = len;
		}
	return(urb);
}

#if defined(__KERNEL24__)
#define _LUsbCompletion(u)  LUsbCompletion(u)
static void LUsbCompletion(struct urb* _urb)
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 19)
#define _LUsbCompletion(u)  LUsbCompletion(u, NULL)
static void LUsbCompletion(struct urb* _urb, struct pt_regs* regs)
#else
#define _LUsbCompletion(u)  LUsbCompletion(u)
static void LUsbCompletion(struct urb* _urb)
#endif
{
	PURB          urb = (PURB)_urb->context;
	int           i;
	unsigned long flags;

	// URB ausketten
	spin_lock_irqsave(&pendingLock, flags);
	LListRemoveEntry(&urb->UrbHeader.Entry);
	spin_unlock_irqrestore(&pendingLock, flags);

	switch (urb->UrbHeader.Function) {
		case URB_FUNCTION_ISOCH_TRANSFER:
//			urb->UrbIsochronousTransfer.TransferBufferLength = _urb->actual_length;
			for (i = 0; i != urb->UrbIsochronousTransfer.NumberOfPackets; i++) {
				urb->UrbIsochronousTransfer.IsoPacket[i].Offset = _urb->iso_frame_desc[i].offset;
				urb->UrbIsochronousTransfer.IsoPacket[i].Length = _urb->iso_frame_desc[i].actual_length;
				urb->UrbIsochronousTransfer.IsoPacket[i].Status = LUsbMapStatus(_urb->iso_frame_desc[i].status, __LINE__);
				}
			break;

		case URB_FUNCTION_BULK_OR_INTERRUPT_TRANSFER:
			urb->UrbBulkOrInterruptTransfer.ActualLength = _urb->actual_length;
			break;

		case URB_FUNCTION_RESET_PIPE:
			_usb_reset_endpoint(_urb->dev, LUsbEndpointFromPipe(urb->UrbPipeRequest.PipeHandle));
			break;

		// ...

		default:
			break;
		}

	LUsbCompleteUrb(urb, LUsbMapStatus(_urb->status, __LINE__));
	_usb_free_urb(_urb);
}

STATUS LINUXAPI LUsbSubmitUrb(void* _dev, PURB urb)
{
	struct usb_device* dev  = (struct usb_device*)_dev;
	struct urb*        _urb = NULL;
	int                i;
	LIST_ENTRY         list, *l1, *l2;
	unsigned long      flags;

	InterlockedIncrement(&urb->UrbHeader.Ref);
	switch (urb->UrbHeader.Function) {
		case URB_FUNCTION_ISOCH_TRANSFER:
			if ((_urb = _usb_alloc_urb(urb->UrbIsochronousTransfer.NumberOfPackets)) != NULL) {
				_urb->pipe                   = urb->UrbIsochronousTransfer.PipeHandle;
				_urb->transfer_buffer        = urb->UrbIsochronousTransfer.TransferBuffer;
				_urb->transfer_buffer_length = urb->UrbIsochronousTransfer.TransferBufferLength;
				_urb->number_of_packets      = urb->UrbIsochronousTransfer.NumberOfPackets;
				_urb->start_frame            = urb->UrbIsochronousTransfer.StartFrame;
				_urb->transfer_flags         = USB_ISO_ASAP;
				_urb->interval               = urb->UrbIsochronousTransfer.Interval;
				for (i = 0; i != urb->UrbIsochronousTransfer.NumberOfPackets; i++) {
					_urb->iso_frame_desc[i].offset        = urb->UrbIsochronousTransfer.IsoPacket[i].Offset;
					_urb->iso_frame_desc[i].length        = urb->UrbIsochronousTransfer.IsoPacket[i].Length;
					_urb->iso_frame_desc[i].actual_length = 0;
					}
				}
			break;

		case URB_FUNCTION_BULK_OR_INTERRUPT_TRANSFER:
			if ((_urb = _usb_alloc_urb(0)) != NULL) {
				_urb->pipe                   = urb->UrbBulkOrInterruptTransfer.PipeHandle;
				_urb->transfer_buffer        = urb->UrbBulkOrInterruptTransfer.TransferBuffer;
				_urb->transfer_buffer_length = urb->UrbBulkOrInterruptTransfer.TransferBufferLength;
//				_urb->number_of_packets      = 0;
				_urb->transfer_flags         = urb->UrbBulkOrInterruptTransfer.TransferFlags & USBD_SHORT_TRANSFER_OK ? 0 : URB_SHORT_NOT_OK;
				_urb->interval               = urb->UrbBulkOrInterruptTransfer.Interval;
				}
			break;

		case URB_FUNCTION_RESET_PIPE:
			_urb = LUsbBuildControlRequest(dev, _usb_sndctrlpipe(dev, 0), USB_RECIP_ENDPOINT, USB_REQ_CLEAR_FEATURE, USB_ENDPOINT_HALT, LUsbEndpointFromPipe(urb->UrbPipeRequest.PipeHandle), NULL, 0);
			break;

		case URB_FUNCTION_ABORT_PIPE:
			LListInit(&list);
			spin_lock_irqsave(&pendingLock, flags);
			for (l1 = pendingUrbs.next; l1 != &pendingUrbs; l1 = l2) {
				struct urb* u;

				l2 = l1->next;
				u  = CONTAINING_RECORD(URB, UrbHeader.Entry, l1)->UrbHeader.Urb;
				if (u->dev == dev && u->pipe == urb->UrbPipeRequest.PipeHandle) {
					LListRemoveEntry(l1);
					LListAddTail(&list, l1);
					}
				}
			while ((l1 = LListRemoveHead(&list)) != NULL) {
				spin_unlock_irqrestore(&pendingLock, flags);
				_usb_unlink_urb(CONTAINING_RECORD(URB, UrbHeader.Entry, l1)->UrbHeader.Urb);
				spin_lock_irqsave(&pendingLock, flags);
				}
			spin_unlock_irqrestore(&pendingLock, flags);
			return(LUsbCompleteUrb(urb, STATUS_SUCCESS));

		case URB_FUNCTION_GET_CURRENT_FRAME_NUMBER:
			urb->UrbGetCurrentFrameNumber.FrameNumber = _usb_get_current_frame_number(dev);
			return(LUsbCompleteUrb(urb, STATUS_SUCCESS));

		case URB_FUNCTION_CONTROL_TRANSFER:
			do {
				unsigned int  pipe;

				pipe = (urb->UrbControlTransfer.RequestType & USB_ENDPOINT_DIRECTION_MASK ? _usb_rcvctrlpipe(dev, 0) : _usb_sndctrlpipe(dev, 0));
				_urb = LUsbBuildControlRequest(dev, pipe, urb->UrbControlTransfer.RequestType, urb->UrbControlTransfer.Request, urb->UrbControlTransfer.Value, urb->UrbControlTransfer.Index, urb->UrbControlTransfer.TransferBuffer, urb->UrbControlTransfer.TransferBufferLength);
				}
			while (0);
			break;

		default:
			// Unbekannte Funktionen direkt completen
			return(LUsbCompleteUrb(urb, STATUS_FAILURE));
		}

	if (_urb == NULL)
		return(LUsbCompleteUrb(urb, STATUS_FAILURE));

	urb->UrbHeader.Urb = _urb;
	spin_lock_irqsave(&pendingLock, flags);
	LListAddTail(&pendingUrbs, &urb->UrbHeader.Entry);
	spin_unlock_irqrestore(&pendingLock, flags);

	_urb->dev           = dev;
	_urb->error_count   = 0;
	_urb->complete      = LUsbCompletion;
	_urb->context       = urb;
	if ((i = _usb_submit_urb(_urb)) == 0)
		return(STATUS_SUCCESS);

	spin_lock_irqsave(&pendingLock, flags);
	LListRemoveEntry(&urb->UrbHeader.Entry);
	spin_unlock_irqrestore(&pendingLock, flags);
	return(LUsbCompleteUrb(urb, LUsbMapStatus(i, __LINE__)));
}

STATUS LINUXAPI LUsbSetInterface(void* dev, int iface, int asetting)
{
   int ret;

   if ((ret = _usb_set_interface((struct usb_device*)dev, iface, asetting)) != 0)
      LinuxDbgPrint(6, "usb_set_interface(iface: %d, asetting: %d) returned %d", iface, asetting, ret);
   return(LUsbMapStatus(ret, __LINE__));
}

STATUS LINUXAPI LUsbStartDevice(void* dev, int iface, int asetting)
{
   return(STATUS_SUCCESS);
}

STATUS LINUXAPI LUsbStopDevice(void* dev)
{
	LIST_ENTRY    list, *l1, *l2;
	struct urb*   u;
	unsigned long flags;

	LListInit(&list);
	spin_lock_irqsave(&pendingLock, flags);
	for (l1 = pendingUrbs.next; l1 != &pendingUrbs; l1 = l2) {
		l2 = l1->next;
		u  = CONTAINING_RECORD(URB, UrbHeader.Entry, l1)->UrbHeader.Urb;
		if (u->dev == dev) {
			LListRemoveEntry(l1);
			LListAddTail(&list, l1);
			}
		}
	while ((l1 = LListRemoveHead(&list)) != NULL) {
		PURB urb = CONTAINING_RECORD(URB, UrbHeader.Entry, l1);

		urb->UrbHeader.Complete = NULL;
		spin_unlock_irqrestore(&pendingLock, flags);
		_usb_unlink_urb(urb->UrbHeader.Urb);
		if (InterlockedDecrement(&urb->UrbHeader.Ref) == 0)
			kfree(urb);
		spin_lock_irqsave(&pendingLock, flags);
		}
	spin_unlock_irqrestore(&pendingLock, flags);
   return(STATUS_SUCCESS);
}

static struct usb_endpoint_descriptor* _UsbFindPipe(void* _dev, int iface, int asetting, unsigned char addr)
{
   struct usb_device* dev = (struct usb_device*)_dev;
   DWORD              idx;

#if defined(__KERNEL24__)
   struct usb_interface_descriptor* setting = &dev->config[0].interface[iface].altsetting[asetting];
   for (idx = 0; idx < setting->bNumEndpoints; idx++) {
      if (setting->endpoint[idx].bEndpointAddress == addr)
         return(&setting->endpoint[idx]);
   	}
#else
	struct usb_host_interface* setting = &dev->config[0].interface[iface]->altsetting[asetting];
   for (idx = 0; idx < setting->desc.bNumEndpoints; idx++) {
      if (setting->endpoint[idx].desc.bEndpointAddress == addr)
         return(&setting->endpoint[idx].desc);
   	}
#endif

   return(NULL);
}

STATUS LINUXAPI LUsbFindPipe(void* dev, int iface, int asetting, unsigned char addr, PUSBD_PIPE_INFORMATION pipe)
{
	struct usb_endpoint_descriptor* ep;

	if ((ep = _UsbFindPipe(dev, iface, asetting, addr)) == NULL)
		return(STATUS_FAILURE);

	pipe->MaximumPacketSize = le16_to_cpu(ep->wMaxPacketSize);
	pipe->EndpointAddress   = ep->bEndpointAddress;
	switch (ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) {
		case USB_ENDPOINT_XFER_CONTROL:
			pipe->PipeType          = UsbdPipeTypeControl;
			pipe->PipeHandle        = usb_sndctrlpipe(dev, addr) | (addr & USB_DIR_IN);
			pipe->Interval          = ep->bInterval;
			break;

		case USB_ENDPOINT_XFER_ISOC:
			pipe->PipeType          = UsbdPipeTypeIsochronous;
			pipe->PipeHandle        = usb_sndisocpipe(dev, addr) | (addr & USB_DIR_IN);
			pipe->Interval          = 1 << (ep->bInterval-1);
			break;

		case USB_ENDPOINT_XFER_BULK:
			pipe->PipeType          = UsbdPipeTypeBulk;
			pipe->PipeHandle        = usb_sndbulkpipe(dev, addr) | (addr & USB_DIR_IN);
			pipe->Interval          = ep->bInterval;
			break;

		case USB_ENDPOINT_XFER_INT:
			pipe->PipeType          = UsbdPipeTypeInterrupt;
			pipe->PipeHandle        = usb_sndintpipe(dev, addr) | (addr & USB_DIR_IN);
			pipe->Interval          = ep->bInterval;
			break;
		}
	return(STATUS_SUCCESS);
}


///////////////////////////////////////////////////////////////////////////////
// HW - PCI stuff
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 3)

static void pci_release_regions(struct pci_dev *pdev)
{
	int i;

	for (i = 0; i < 6; i++) {
		if (pci_resource_len(pdev, i) == 0)
			continue;

		if (pci_resource_flags(pdev, i) & IORESOURCE_IO)
			release_region(pci_resource_start(pdev, i), pci_resource_len(pdev, i));
		else if (pci_resource_flags(pdev, i) & IORESOURCE_MEM)
			release_mem_region(pci_resource_start(pdev, i), pci_resource_len(pdev, i));
		}
}

static int pci_request_regions(struct pci_dev *pdev, char *res_name)
{
	int i;

	for (i = 0; i < 6; i++) {
		if (pci_resource_len(pdev, i) == 0)
			continue;

		if (pci_resource_flags(pdev, i) & IORESOURCE_IO) {
			if (!request_region(pci_resource_start(pdev, i), pci_resource_len(pdev, i), res_name))
				goto err_out;
			}
		else if (pci_resource_flags(pdev, i) & IORESOURCE_MEM) {
			if (!request_mem_region(pci_resource_start(pdev, i), pci_resource_len(pdev, i), res_name))
				goto err_out;
			}
		}

	return(0);

err_out:
	printk (KERN_WARNING "PCI: Unable to reserve %s region #%d:%lx@%lx for device %s\n", pci_resource_flags(pdev, i) & IORESOURCE_IO ? "I/O" : "mem", i + 1, /* PCI BAR # */ pci_resource_len(pdev, i), pci_resource_start(pdev, i), pdev->slot_name);
	pci_release_regions(pdev);
	return(-EBUSY);
}

#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 4)

static void pci_disable_device(struct pci_dev *dev)
{
	u16 pci_command;

	pci_read_config_word(dev, PCI_COMMAND, &pci_command);
	if (pci_command & PCI_COMMAND_MASTER) {
		pci_command &= ~PCI_COMMAND_MASTER;
		pci_write_config_word(dev, PCI_COMMAND, pci_command);
		}
}

#endif

static int OS_PciProbe(struct pci_dev* dev, const struct pci_device_id* id)
{
	int        i;
	HARDW_DEV  hw;
	LPWORD     pPort;
	LPDWORD    pMem;
	LPBYTE*    pVaddr;
	ULONG      io_flags;
	LPVOID     ptr;

	if (pci_enable_device(dev)) {
		LinuxDbgPrint(6, "Enabling PCI device failed");
		return(-1);
		}

	pci_set_master(dev);
	if (pci_request_regions(dev, (char*)__drvName)) {
		LinuxDbgPrint(6, "PCI I/O register failed");
		return(-1);
		}

	memset(&hw, 0, sizeof(hw));
	hw.devParms.irq = dev->irq;
	pPort  = &hw.devParms.port1;
	pMem   = &hw.devParms.mem1;
	pVaddr = &hw.devParms.ptr1;
	for (i = 0; i < 6; i++) {
		io_flags = pci_resource_flags(dev, i);
		if ((io_flags & IORESOURCE_IO) != 0 && pPort <= &hw.devParms.port2)
			*pPort++ = pci_resource_start(dev, i);
		else if ((io_flags & IORESOURCE_MEM) != 0 && pMem <= &hw.devParms.mem3) {
			*pMem     = pci_resource_start(dev, i);
			*pVaddr++ = (LPBYTE)ioremap(*pMem++, pci_resource_len(dev, i));
			}
		}

	hw.devID      = dev->bus->number + 256*dev->devfn;
	hw.devAddr[0] = (DWORD)dev->device + 65536*(DWORD)dev->vendor;
	hw.devAddr[1] = (DWORD)dev->subsystem_vendor + 65536*(DWORD)dev->subsystem_device;
	hw.pModel     = (PSUPHWDEV)id->driver_data;
	hw.pOsDev     = dev;
	if ((ptr = DriverAddDevice(&hw)) == NULL) {
		DbgTrace("PCI_PROBE: DriverAddDevice failed");
		pci_release_regions(dev);
		pci_disable_device(dev);
		return(-1);
		}

	pci_set_drvdata(dev, AllocHWData(ptr, &hw));
	return(0);
}

static void OS_PciRemove(struct pci_dev* dev)
{
	PRIV_DATA* phw;
	int        i;

	if ((phw = (PRIV_DATA*)pci_get_drvdata(dev)) != NULL) {
		pci_set_drvdata(dev, NULL);
		if (inExit == 0)
			ExecuteIrp(NULL, pci_get_drvdata(dev), IRP_MJ_PNP, IRP_MN_REMOVE_DEVICE, NULL, 0, NULL, 0);
		for (i = 0; i < 3 && phw->vptr[i] != NULL; i++)
			iounmap(phw->vptr[i]);
		FreeHWData(phw);
		pci_release_regions(dev);
		pci_disable_device(dev);
		}
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 12)

static int OS_PciSuspend(struct pci_dev* dev, pm_message_t state)
{
	int res;

	if ((res = ExecuteIrp(NULL, pci_get_drvdata(dev), IRP_MJ_POWER, IRP_ARG_POWER_DOWN, NULL, 0, NULL, 0)) < 0)
		res = -EBUSY;
	return(res);
}

static int OS_PciResume(struct pci_dev* dev)
{
	int res;

	if ((res = ExecuteIrp(NULL, pci_get_drvdata(dev), IRP_MJ_POWER, IRP_ARG_POWER_UP, NULL, 0, NULL, 0)) < 0)
		res = -EBUSY;
	return(res);
}

#endif

static struct pci_driver OS_pciDriver = {
   name:           NULL,
   id_table:       NULL,
   probe:          OS_PciProbe,
   remove:         OS_PciRemove,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 12)
	suspend:        OS_PciSuspend,
	resume:         OS_PciResume,
#endif
	};

static struct pci_device_id* OS_pciTable;

static void LPciUnregisterDrv(LPVOID ptr)
{
	LinuxDbgPrint(6, "Unregistering PCI driver...");
	pci_unregister_driver(&OS_pciDriver);
	HeapFree(OS_pciTable);
}

static int LINUXAPI LPciRegisterDrv(struct pci_driver* drv)
{
   int result;

	LinuxDbgPrint(6, "Registering PCI driver...");
   if ((result = pci_register_driver(drv)) < 0) {
      LinuxDbgPrint(6, "   pci_register_driver returned %D", result);
		HeapFree(OS_pciTable);
      return -1;
	   }
	atExit(LPciUnregisterDrv, NULL);
   return 0;
}

void LINUXAPI LPciConfigWrite(LPVOID pDev, LPVOID buf, DWORD ofs, DWORD len)
{
   struct pci_dev* dev = (struct pci_dev*)pDev;

   if (len == 1)
      pci_write_config_byte(dev, (int)ofs, *(u8*)buf);
   else if (len == 2)
      pci_write_config_word(dev, (int)ofs, *(u16*)buf);
   else if (len <= 4 )
      pci_write_config_dword(dev, (int)ofs, *(u32*)buf);
}

void LINUXAPI LPciConfigRead(LPVOID pDev, LPVOID buf, DWORD ofs, DWORD len)
{
   struct pci_dev* dev = (struct pci_dev*)pDev;

   if (len == 1)
      pci_read_config_byte(dev,  (int)ofs, (u8*)buf);
   else if (len == 2)
      pci_read_config_word(dev,  (int)ofs, (u16*)buf);
   else if (len <= 4 )
      pci_read_config_dword(dev, (int)ofs, (u32*)buf);
}

///////////////////////////////////////////////////////////////////////////////
// HW - Powermanagement

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 12)

static int OS_PmCallback(struct pm_dev* pmHdl, pm_request_t rqst, LPVOID data)
{
	int res;

	if ((res = ExecuteIrp(NULL, pmHdl->data, IRP_MJ_POWER, (rqst == PM_SUSPEND ? IRP_ARG_POWER_DOWN : IRP_ARG_POWER_UP), NULL, 0, NULL, 0)) < 0)
		res = -EBUSY;
	return(res);
}

LPVOID LINUXAPI LPmRegisterDev(LPVOID irpHdl, DWORD devType, DWORD id)
{
	pm_dev_t       type;
   struct pm_dev* pmHdl;

   switch (devType) {
      case DEV_TYPE_PCI:
         type = PM_PCI_DEV;
         break;

		case DEV_TYPE_USB:
         type = PM_USB_DEV;
         break;

		case DEV_TYPE_ISA:
         type = PM_ISA_DEV;
         break;

		default:
         type = PM_UNKNOWN_DEV;
         break;
   	}
   if ((pmHdl = pm_register(type, id, OS_PmCallback)) != NULL)
   	pmHdl->data = AllocHWData(irpHdl, NULL);
   return(pmHdl);
}

void LINUXAPI LPmUnregisterDev(LPVOID pmHdl)
{
	FreeHWData(((struct pm_dev*)pmHdl)->data);
	pm_unregister((struct pm_dev*)pmHdl);
}

#else

LPVOID LINUXAPI LPmRegisterDev(LPVOID irpHdl, DWORD devType, DWORD id)
{
   return(NULL);
}

void LINUXAPI LPmUnregisterDev(LPVOID pmHdl)
{
}

#endif

///////////////////////////////////////////////////////////////////////////////
// HW - Hardware device init and exit stuff

STATUS LINUXAPI LRegSupHwDev(const PSUPHWDEV dev_tab, LPCSTR drv_name)
{
	int    pci, usb;
	struct pci_device_id* pPtr;
	struct usb_device_id* uPtr;

	STATUS ret    = STATUS_SUCCESS;
	PSUPHWDEV dev = (PSUPHWDEV)dev_tab;

	for (pci = usb = 0; dev->type != 0; dev++) {
		if (dev->type == DEV_TYPE_PCI)
			pci++;
		else if (dev->type == DEV_TYPE_USB)
			usb++;
		}

	if (pci != 0) {
		pPtr        = HeapAllocate((pci+1) * sizeof(struct pci_device_id));
		OS_pciTable = pPtr;
		for (dev = dev_tab; dev->type != 0; dev++) {
			if (dev->type == DEV_TYPE_PCI) {
				pPtr->vendor      = dev->vendor;
				pPtr->device      = (dev->device    ? dev->device    : (UINT)PCI_ANY_ID);
				pPtr->subvendor   = (dev->subvendor ? dev->subvendor : (UINT)PCI_ANY_ID);
				pPtr->subdevice   = (dev->subdevice ? dev->subdevice : (UINT)PCI_ANY_ID);
				pPtr->class       = 0;
				pPtr->class_mask  = 0;
				pPtr->driver_data = (unsigned long)dev;
				pPtr++;
				}
			}
		memset(pPtr, 0, sizeof(struct pci_device_id));
		OS_pciDriver.name     = (char*)drv_name;
		OS_pciDriver.id_table = OS_pciTable;
		if (LPciRegisterDrv(&OS_pciDriver) < 0)
			ret |= ERROR_REG_PCI;
		}

	if (usb) {
		uPtr        = HeapAllocate((usb+1) * sizeof(struct usb_device_id));
		OS_usbTable = uPtr;
		for (dev = dev_tab; dev->type != 0; dev++) {
			if (dev->type == DEV_TYPE_USB) {
				uPtr->match_flags     = USB_DEVICE_ID_MATCH_VENDOR|USB_DEVICE_ID_MATCH_PRODUCT|(dev->subvendor ? USB_DEVICE_ID_MATCH_INT_CLASS : 0);
				uPtr->idVendor        = dev->vendor;
				uPtr->idProduct       = dev->device;
				uPtr->bInterfaceClass = dev->subvendor;
				uPtr->driver_info     = (unsigned long)dev;
				uPtr++;
				}
			}
		memset(uPtr, 0, sizeof(struct usb_device_id));
		OS_usbDriver.name     = drv_name;
		OS_usbDriver.id_table = OS_usbTable;
		if (LUsbRegisterDrv(&OS_usbDriver) < 0)
			ret |= ERROR_REG_USB;
		}

	return(ret);
}

STATUS LINUXAPI LUnregSupHwDev(void)
{
	doExit();
	return(STATUS_SUCCESS);
}

///////////////////////////////////////////////////////////////////////////////
// Persistent Storage
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 39)
#define __LOCK_KERNEL()   lock_kernel()
#define __UNLOCK_KERNEL() unlock_kernel()
#else
#define __LOCK_KERNEL()   do {} while (0)
#define __UNLOCK_KERNEL() do {} while (0)
#endif

#define LOCK_KERNEL()                                        \
	do {                                                     \
		mm_segment_t fs;                                     \
		__LOCK_KERNEL();                                     \
		fs = get_fs();                                       \
		set_fs(get_ds())

#define UNLOCK_KERNEL()                                      \
		smp_mb();                                            \
		set_fs(fs);                                          \
		__UNLOCK_KERNEL();                                   \
		}                                                    \
	while (0)

#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0)
#define __WRITE(f, buf, len)                                 \
	vfs_write(f, buf, len, &f->f_pos)
#define __READ(f, buf, len)                                  \
	vfs_read(f, buf, len, &f->f_pos)
#else
#define __WRITE(f, buf, len)                                 \
	kernel_write(f, buf, len, &f->f_pos)
#define __READ(f, buf, len)                                  \
	kernel_read(f, buf, len, &f->f_pos)
#endif

void LINUXAPI LSetPersist(LPCSTR key, LPVOID buf, int len)
{
	struct file* f;

	LOCK_KERNEL();
	f = filp_open(key, O_CREAT|O_TRUNC|O_WRONLY, 0x180);
	if (!IS_ERR(f)) {
		__WRITE(f, buf, len);
		filp_close(f, NULL);
		}
	UNLOCK_KERNEL();
}

int LINUXAPI LGetPersist(LPCSTR key, LPVOID buf, int len)
{
	int          res = -1;
	struct file* f;

	LOCK_KERNEL();
	f = filp_open(key, O_RDONLY, 0x180);
	if (!IS_ERR(f)) {
		res = __READ(f, buf, len);
		filp_close(f, NULL);
		}
	UNLOCK_KERNEL();
	return(res);
}

int LINUXAPI LDumpFile(LPCSTR fnam, LPVOID buf, int len)
{
	struct file* f;

	LOCK_KERNEL();
	if (buf == NULL) {
		if (!IS_ERR(f = filp_open(fnam, O_CREAT|O_TRUNC|O_WRONLY, 0x180)))
			filp_close(f, NULL);
		}
	else if (!IS_ERR(f = filp_open(fnam, O_CREAT|O_WRONLY, 0x180))) {
		__WRITE(f, buf, len);
		filp_close(f, NULL);
		}
	UNLOCK_KERNEL();
	return(!IS_ERR(f));
}

///////////////////////////////////////////////////////////////////////////////
// Memory Pool
typedef struct {
	void*     list;
	void*     next;
	void*     first;
	void*     top;
	size_t    siz;
	} LMEMPOOL;

static void* LMemPoolAlloc(LMEMPOOL* pool)
{
	void* res;
	DWORD cpu;

	cpu = LSMPLock();
	if ((res = pool->list) != NULL)
		pool->list = *(void**)res;
	else if ((res = pool->next) < pool->top)
		pool->next = (unsigned char*)res + pool->siz;
	else
		res = NULL;
	LSMPUnlock(cpu);
	return(res ? (unsigned char*)res + sizeof(void*) : res);
}

static void LMemPoolFree(LMEMPOOL* pool, void* ptr)
{
	DWORD cpu;
	void* pxx;

	if (ptr != NULL) {
		ptr = (unsigned char*)ptr - sizeof(void*);
		cpu = LSMPLock();

		__ASSERT(ptr >= pool->first && ptr < pool->top);
		__ASSERT(((unsigned char*)ptr - (unsigned char*)pool->first) % pool->siz == 0);
		for (pxx = pool->list; pxx != NULL; pxx = *(void**)pxx)
			__ASSERT(pxx != ptr);

		*(void**)ptr = pool->list;
		pool->list   = ptr;
		LSMPUnlock(cpu);
		}
}

#define DEFINE_MEM_POOL(nam, typ, elems) \
	static unsigned char __mem ## nam[(elems)*(sizeof(void*) + sizeof(typ))]; \
	static LMEMPOOL nam = { NULL, __mem ## nam, __mem ## nam, &__mem ## nam[(elems)*(sizeof(void*) + sizeof(typ))], sizeof(void*) + sizeof(typ) }

///////////////////////////////////////////////////////////////////////////////
// Waiter API
#define LWAITABLE_MUTEX    0
#define LWAITABLE_EVENT    1
#define LWAITABLE_THREAD   2

struct _GENERIC {
	};

struct _MUTEX {
	void*               owner;
	};

struct _THREAD {
	LIST_ENTRY          list;
	void*               ctx;
	LTHREADFUNC         func;
	};

struct _SPINLOCK {
	spinlock_t          lock;
	};

struct _DPC {
	LIST_ENTRY          list;
	void*               ctx;
	LDPCFUNC            func;
	};

typedef struct {
	wait_queue_head_t   queue;
	unsigned short      type;
	unsigned short      flag;
	union {
		struct _MUTEX       mutex;
		struct _THREAD      thread;
		struct _SPINLOCK    spinlock;
		struct _DPC         dpc;
		};
	} LWAITABLE;

DEFINE_MEM_POOL(poolWait, LWAITABLE, 128);

LWAITABLE* LWaitableInit(unsigned short type)
{
	LWAITABLE* wait;

	if ((wait = (LWAITABLE*)LMemPoolAlloc(&poolWait)) != NULL) {
		init_waitqueue_head(&wait->queue);
		wait->type = type;
		wait->flag = 0;
		}
	return(wait);
}

void LINUXAPI LWaitableDestroy(void* obj)
{
	LMemPoolFree(&poolWait, obj);
}

int LINUXAPI LWaitableWait(void* obj, long timeout)
{
	LWAITABLE* wait = (LWAITABLE*)obj;
	DWORD      cpu;

	cpu = LSMPLock();
	switch (wait->type) {
		case LWAITABLE_MUTEX:
			if (wait->mutex.owner == current) {
				wait->flag += 1;
				LSMPUnlock(cpu);
				return(1);
				}
			break;

		case LWAITABLE_THREAD:
			if (wait->flag)
				wait->flag = 2;
			break;

		// ...
		}
	timeout = (timeout < 0 ? MAX_SCHEDULE_TIMEOUT : ms2jif(timeout));
	while (1) {
		LSMPUnlock(cpu);
		timeout = wait_event_interruptible_timeout(wait->queue, wait->flag == 0, timeout);
		// BUGBUGBUG: Was ist mit ERESTARTSYS???
		cpu = LSMPLock();
		if (wait->flag == 0) {
			switch (wait->type) {
				case LWAITABLE_MUTEX:
					wait->mutex.owner = current;
					wait->flag = 1;
					break;

				case LWAITABLE_EVENT:
					wait->flag = 1;
					break;

				// ...
				}
			LSMPUnlock(cpu);
			return(1);
			}
		if (timeout <= 0) {
			LSMPUnlock(cpu);
			return(0);
			}
		}
}

void LINUXAPI LWaitableWake(void* obj)
{
	LWAITABLE*  wait = (LWAITABLE*)obj;
	DWORD       cpu, cnt;

	cpu = LSMPLock();
	cnt = 1;
	switch (wait->type) {
		case LWAITABLE_MUTEX:
			if (--wait->flag != 0) {
				LSMPUnlock(cpu);
				return;
				}
			wait->mutex.owner = NULL;
			break;

		case LWAITABLE_EVENT:
			cnt = 0;
			// FALL-THRU

		default:
			wait->flag = 0;
			break;
		}
	wake_up_interruptible_nr(&wait->queue, cnt);
	LSMPUnlock(cpu);
}

///////////////////////////////////////////////////////////////////////////////
// Mutex API
LMUTEX LINUXAPI LMutexInit(void)
{
	LWAITABLE* lock;

	if ((lock = LWaitableInit(LWAITABLE_MUTEX)) != NULL)
		lock->mutex.owner = NULL;
	return((LMUTEX)lock);
}

///////////////////////////////////////////////////////////////////////////////
// Event API
LEVENT LINUXAPI LEventInit(void)
{
	LWAITABLE* evt;

	if ((evt = LWaitableInit(LWAITABLE_EVENT)) != NULL)
		evt->flag = 1;
	return((LEVENT)evt);
}

void LINUXAPI LEventClear(LEVENT evt)
{
	DWORD cpu;

	cpu = LSMPLock();
	((LWAITABLE*)evt)->flag = 1;
	LSMPUnlock(cpu);
}

////////////////////////////////////
// Thread API
#if defined(__KERNEL24__)
#define kthread_run(a, b, c)   kernel_thread(a, b, 0)
#else
#include <linux/kthread.h>
#endif

static int LThreadLauncher(void* arg)
{
	LWAITABLE* thrd = (LWAITABLE*)arg;

	set_user_nice(current, -19);
	(*thrd->thread.func)((LTHREAD)thrd, thrd->thread.ctx);
	LWaitableWake(thrd);
	return(0);
}

LTHREAD LINUXAPI LThreadInit(LTHREADFUNC func, void* ctx)
{
	LWAITABLE* thrd;

	if ((thrd = LWaitableInit(LWAITABLE_THREAD)) != NULL) {
		thrd->flag         = 1;
		thrd->thread.ctx   = ctx;
		thrd->thread.func  = func;
		kthread_run(LThreadLauncher, thrd, "kal_thread");
		}
	return((LTHREAD)thrd);
}

///////////////////////////////////////////////////////////////////////////////
// Spinlock API
// BUGBUGBUG: Linux spin_lock verwenden...
LSPINLOCK LINUXAPI LSpinlockInit(void)
{
	LWAITABLE* lock;

	if ((lock = (LWAITABLE*)LMemPoolAlloc(&poolWait)) != NULL)
		spin_lock_init(&lock->spinlock.lock);
	return((LSPINLOCK)lock);
}

void LINUXAPI LSpinlockAcquire(LSPINLOCK lock)
{
	spin_lock(&((LWAITABLE*)lock)->spinlock.lock);
}

void LINUXAPI LSpinlockRelease(LSPINLOCK lock)
{
	spin_unlock(&((LWAITABLE*)lock)->spinlock.lock);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DPC API
static LIST_ENTRY dpcQueue = {&dpcQueue, &dpcQueue};
static void       dpcHandler(unsigned long ctx);
static DECLARE_TASKLET(dpcTask, dpcHandler, 0);

static void dpcHandler(unsigned long ctx)
{
	PLIST_ENTRY ent;
	LWAITABLE*  dpc;
	DWORD       cpu;

	cpu = LSMPLock();
	while ((ent = LListRemoveHead(&dpcQueue)) != NULL) {
		LSMPUnlock(cpu);
		dpc = CONTAINING_RECORD(LWAITABLE, dpc.list, ent);
		(*dpc->dpc.func)((LDPC)dpc, dpc->dpc.ctx);
		cpu = LSMPLock();
		}
	LSMPUnlock(cpu);
}


LDPC LINUXAPI LDpcInit(LDPCFUNC func, LPVOID ctx)
{
	LWAITABLE* dpc;

	if ((dpc = (LWAITABLE*)LMemPoolAlloc(&poolWait)) != NULL) {
		LListInit(&dpc->dpc.list);
		dpc->dpc.ctx  = ctx;
		dpc->dpc.func = func;
		}
	return((LDPC)dpc);
}

void LINUXAPI LDpcQueue(LDPC _dpc)
{
	LWAITABLE* dpc = (LWAITABLE*)_dpc;
	DWORD      cpu;

	cpu = LSMPLock();
	if (LListIsEmpty(&dpc->dpc.list)) {
		LListAddTail(&dpcQueue, &dpc->dpc.list);
		tasklet_schedule(&dpcTask);
		}
	LSMPUnlock(cpu);
}

///////////////////////////////////////////////////////////////////////////////
// Socket API
#include <linux/net.h>
#include <linux/utsname.h>
#include <net/sock.h>
#include <net/tcp_states.h>

#if defined(__KERNEL24__)
#define SOCK_ELEM(sk, elem)     sk->elem
#elif defined(__KERNEL26__)
#define SOCK_ELEM(sk, elem)     sk->sk_ ## elem
#endif

#define SOCKET_ELEM(_sk, _elem) SOCK_ELEM(_sk->sk, _elem)

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 15, 0)
static void _sock_data_ready(struct sock* sk, int bytes)
#else
static void _sock_data_ready(struct sock* sk)
#endif
{
	LSOCKET* sock = (LSOCKET*)SOCK_ELEM(sk, user_data);

	if (sock != NULL && sock->onDataReady != NULL)
		(*sock->onDataReady)(sock, sock->context);
}

static void _sock_connected(struct sock* sk)
{
	LSOCKET* sock = (LSOCKET*)SOCK_ELEM(sk, user_data);

	if (SOCK_ELEM(sk, state) == TCP_ESTABLISHED && sock != NULL && sock->onDataReady != NULL)
		(*sock->onDataReady)(sock, sock->context);
}

static void _sock_state_change(struct sock* sk)
{
	LSOCKET* sock = (LSOCKET*)SOCK_ELEM(sk, user_data);

	if ((SOCK_ELEM(sk, shutdown) & RCV_SHUTDOWN) != 0 && sock != NULL && sock->onDisconnect != NULL) {
		SOCK_ELEM(sk, state_change) = sock->state_change;
		(*sock->onDisconnect)(sock, sock->context);
		}
}

// static struct socket* _sockFree[32];
// static int            _sockWr = 0, _sockRd = 0;

// static void __WORK_HANDLER(__sock_release)
// {
// 	DWORD          cpu;
// 	struct socket* sk;

// 	cpu = LSMPLock();
// 	while (_sockRd != _sockWr) {
// 		sk = _sockFree[_sockRd++];
// 		if (_sockRd == sizeof(_sockFree)/sizeof(_sockFree[0]))
// 			_sockRd = 0;
// 		// LSMPUnlock(cpu);
// 		// sock_release(sk);
// 		// cpu = LSMPLock();
// 		}
// 	LSMPUnlock(cpu);
// }

static void _sock_setup(LSOCKET* sock, struct socket* sk)
{
	sock->sock                 = sk;
	sock->data_ready           = SOCKET_ELEM(sk, data_ready);
	sock->state_change         = SOCKET_ELEM(sk, state_change);
	SOCKET_ELEM(sk, user_data) = sock;
	SOCKET_ELEM(sk, reuse)     = 1;
}

static void _sock_release(struct socket* sk)
{
	// static __WORK_STRUCT(releaseTask, __sock_release);

	// DWORD  cpu;

	SOCKET_ELEM(sk, user_data) = NULL;
	sock_release(sk);
	// cpu = LSMPLock();
	// _sockFree[_sockWr++] = sk;
	// if (_sockWr == sizeof(_sockFree)/sizeof(_sockFree[0]))
	// 	_sockWr = 0;
	// LSMPUnlock(cpu);
	// __SCHEDULE_WORK(&releaseTask);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 2, 0)
	#define _sock_create_kern(a, b, c, d) sock_create_kern(&init_net, a, b, c, d)
#else
	#define _sock_create_kern(a, b, c, d) sock_create_kern(a, b, c, d)
#endif

int LINUXAPI LSocketCreate(LSOCKET* sock, int domain, int type, int protocol)
{
	struct socket* sk = NULL;
	int            res;

	sock->context      = NULL;
	sock->onDataReady  = NULL;
	sock->onDisconnect = NULL;
	sock->sock         = NULL;

	if ((res = _sock_create_kern(domain, type, protocol, &sk)) >= 0)
		_sock_setup(sock, sk);
	return res;
}

int LINUXAPI LSocketBind(LSOCKET* sock, LSOCKADDR* addr)
{
	struct socket* sk = (struct socket*)sock->sock;

	SOCKET_ELEM(sk, data_ready)   = _sock_data_ready;
	SOCKET_ELEM(sk, state_change) = _sock_state_change;
	return kernel_bind(sk, (struct sockaddr*)addr, sizeof(LSOCKADDR));
}

int LINUXAPI LSocketListen(LSOCKET* sock, int backlog)
{
	struct socket* sk = (struct socket*)sock->sock;

	if ((unsigned)backlog > SOMAXCONN)
		backlog = SOMAXCONN;
	SOCKET_ELEM(sk, state_change) = _sock_connected;
	return kernel_listen(sk, backlog);
}

int LINUXAPI LSocketConnect(LSOCKET* sock, LSOCKADDR* addr)
{
	struct socket* sk = (struct socket*)sock->sock;

	SOCKET_ELEM(sk, data_ready)   = _sock_data_ready;
	SOCKET_ELEM(sk, state_change) = _sock_state_change;
	return kernel_connect(sk, (struct sockaddr*)addr, sizeof(LSOCKADDR), 0);
}

int LINUXAPI LSocketAccept(LSOCKET* sock, LSOCKET* parent)
{
	struct socket* sk  = (struct socket*)parent->sock;
	struct socket* ask;
	int            res;

	sock->context      = NULL;
	sock->onDataReady  = NULL;
	sock->onDisconnect = NULL;
	sock->sock         = NULL;
	if ((res = kernel_accept(sk, &ask, 0)) >= 0) {
		_sock_setup(sock, ask);
		SOCKET_ELEM(ask, data_ready)   = _sock_data_ready;
		SOCKET_ELEM(ask, state_change) = _sock_state_change;
		}
	return res;
}

int LINUXAPI LSocketRecvFrom(LSOCKET* sock, void* buf, unsigned int len, LSOCKADDR* addr)
{
	struct socket* sk  = (struct socket*)sock->sock;
	int            res = 0;

	if (sk) {
		struct msghdr  msg;
		struct kvec    iov;

		iov.iov_base       = buf;
		iov.iov_len        = len;

		msg.msg_name       = (struct sockaddr*)addr;
		msg.msg_namelen    = (addr == NULL ? 0 : sizeof(LSOCKADDR));
		msg.msg_control    = NULL;
		msg.msg_controllen = 0;
		msg.msg_flags      = MSG_DONTWAIT;

		if ((res = kernel_recvmsg(sk, &msg, &iov, 1, len, MSG_DONTWAIT)) < 0)
			res = 0;
		}
	return(res);
}

int LINUXAPI LSocketRecv(LSOCKET* sock, void* buf, unsigned int len)
{
	return(LSocketRecvFrom(sock, buf, len, NULL));
}

int LINUXAPI LSocketSendTo(LSOCKET* sock, void* buf, unsigned int len, LSOCKADDR* addr)
{
	struct socket* sk  = (struct socket*)sock->sock;
	int            res = 0;

	if (sk) {
		struct msghdr  msg;
		struct kvec    iov;

		iov.iov_base       = buf;
		iov.iov_len        = len;

		msg.msg_name       = addr;
		msg.msg_namelen    = (addr == NULL ? 0 : sizeof(LSOCKADDR));
		msg.msg_control    = NULL;
		msg.msg_controllen = 0;
		msg.msg_flags      = MSG_DONTWAIT;

		if ((res = kernel_sendmsg(sk, &msg, &iov, 1, len)) < 0)
			res = 0;
		}
	return(res);
}

int LINUXAPI LSocketSend(LSOCKET* sock, void* buf, unsigned int len)
{
	return(LSocketSendTo(sock, buf, len, NULL));
}

int LINUXAPI LSocketGetSockName(LSOCKET* sock, LSOCKADDR* addr)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 17, 0)
	int len = sizeof(LSOCKADDR);

	return kernel_getsockname((struct socket*)sock->sock, (struct sockaddr*)addr, &len);
#else
	return kernel_getsockname((struct socket*)sock->sock, (struct sockaddr*)addr);
#endif
}

int LINUXAPI LSocketGetPeerName(LSOCKET* sock, LSOCKADDR* addr)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 17, 0)
	int len = sizeof(LSOCKADDR);

	return kernel_getpeername((struct socket*)sock->sock, (struct sockaddr*)addr, &len);
#else
	return kernel_getpeername((struct socket*)sock->sock, (struct sockaddr*)addr);
#endif
}

int LINUXAPI LSocketSetSockOpt(LSOCKET* sock, int level, int optname, char* optval, unsigned int optlen)
{
	return kernel_setsockopt((struct socket*)sock->sock, level, optname, optval, optlen);
}

int LINUXAPI LSocketGetSockOpt(LSOCKET* sock, int level, int optname, char* optval, int* optlen)
{
	return kernel_getsockopt((struct socket*)sock->sock, level, optname, optval, optlen);
}

int LINUXAPI LSocketShutdown(LSOCKET* sock, int flags)
{
	return kernel_sock_shutdown((struct socket*)sock->sock, flags);
}

void LINUXAPI LSocketClose(LSOCKET* sock)
{
	struct socket* sk  = (struct socket*)sock->sock;

	if (sk != NULL) {
		sock->sock = NULL;
		sock->context = NULL;
		SOCKET_ELEM(sk, data_ready)   = sock->data_ready;
		SOCKET_ELEM(sk, state_change) = sock->state_change;
		_sock_release(sk);
		}
}

///////////////////////////////////////////////////////////////////////////////
// Network Notifier Stuff
#include <linux/notifier.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/if_arp.h>

static int _netdevice_count = 0;

extern void LINUXAPI LInetAddrNotifier(unsigned int flag, DWORD addr, char* guid, unsigned char* hwaddr, unsigned int hwaddr_len);

static int _netdevice_notifier(struct notifier_block* self, unsigned long event, void* parm)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 12, 0)
	struct net_device *dev = netdev_notifier_info_to_dev(parm);
#else
	struct net_device *dev = (struct net_device*)parm;
#endif
	struct in_device*  in_dev = (struct in_device*)dev->ip_ptr;
	struct in_ifaddr*  ifa;

	if (_netdevice_count && in_dev != NULL && dev->type != ARPHRD_LOOPBACK) {
		switch (event) {
			case NETDEV_UP:
			case NETDEV_DOWN:
				for (ifa = in_dev->ifa_list; ifa /* && !(ifa->ifa_flags & IFA_F_SECONDARY) */; ifa = ifa->ifa_next)
					LInetAddrNotifier(event == NETDEV_UP, ifa->ifa_address, dev->name, dev->dev_addr, dev->addr_len);
				break;
			}
		}
	return(NOTIFY_DONE);
}

static int _inetaddr_notifier(struct notifier_block* self, unsigned long event, void* parm)
{
	struct in_ifaddr*  addr = (struct in_ifaddr*)parm;
	struct net_device* dev  = addr->ifa_dev->dev;

	if (_netdevice_count)
		LInetAddrNotifier(event == NETDEV_UP, addr->ifa_address, dev->name, dev->dev_addr, dev->addr_len);
	return(NOTIFY_DONE);
}

static struct notifier_block _netdevice_block = {_netdevice_notifier, NULL, 0};
static struct notifier_block _inetaddr_block  = {_inetaddr_notifier, NULL, 0};

void LINUXAPI LRegisterNetworkNotifier(void)
{
	if (_netdevice_count++ == 0) {
		register_netdevice_notifier(&_netdevice_block);
		register_inetaddr_notifier(&_inetaddr_block);
		}
}

void LINUXAPI LUnregisterNetworkNotifier(void)
{
	if (_netdevice_count != 0 && --_netdevice_count == 0) {
		unregister_netdevice_notifier(&_netdevice_block);
		unregister_inetaddr_notifier(&_inetaddr_block);
		}
}

///////////////////////////////////////////////////////////////////////////////
// FloatingPointStuff
// #include <asm/i387.h>

void LINUXAPI LSaveFloatingPointState(void)
{
//	kernel_fpu_begin();
}

void LINUXAPI LRestoreFloatingPointState(void)
{
//	kernel_fpu_end();
}

///////////////////////////////////////////////////////////////////////////////
// Performancecounter Stuff
#include "asm/msr.h"

#if !defined(rdtscl)
#define rdtscl(v)  ((v) = (u32)rdtsc_ordered())
#endif

#if !defined(rdtscll)
#define rdtscll(v)  ((v) = rdtsc_ordered())
#endif

static inline DWORD getusec(void)
{
	struct timeval tv;
	do_gettimeofday(&tv);
	return((DWORD)tv.tv_usec);
}

void LINUXAPI LQueryPerformanceFrequency(DWORD* ticks, DWORD* incr)
{
	DWORD t1, t2;
	DWORD start, stop;

	// Taktrate ermitteln
	t1 = getusec();
	do {} while (t1 == getusec());
	rdtscl(start);
	t1 = getusec();
	do {} while (t1 == (t2 = getusec()));
	rdtscl(stop);
	*ticks = stop - start;
	*incr  = t2 - t1;
}

void LINUXAPI LQueryPerformanceCounter(unsigned long long int* val)
{
	rdtscll(*val);
}

///////////////////////////////////////////////////////////////////////////////
// Linux Kernel Mode CAPI Interface
#include <linux/skbuff.h>
#include <linux/capi.h>

typedef void (*SIGNALFUNC)(WORD, LPVOID);
typedef void (*MYSIGNALFUNC)(WORD, LPVOID) LINUXAPI;

extern WORD   LINUXAPI capi_register(capi_register_params* parms, LPWORD applid);
extern WORD   LINUXAPI capi_release(WORD applid);
extern WORD   LINUXAPI capi_put_message_helper(WORD applid, LPVOID msg);
extern WORD   LINUXAPI capi_get_message_helper(WORD applid, LPVOID* msg);
extern WORD   LINUXAPI capi_set_signal(WORD applid, MYSIGNALFUNC signal, LPVOID parm);
extern WORD   LINUXAPI capi_get_manufacturer(UINT ctrl, BYTE buf[CAPI_MANUFACTURER_LEN]);
extern WORD   LINUXAPI capi_get_version(UINT ctrl, struct capi_version* ver);
extern WORD   LINUXAPI capi_get_serial(UINT ctrl, BYTE serial[CAPI_SERIAL_LEN]);
extern WORD   LINUXAPI capi_get_profile(UINT ctrl, struct capi_profile* prof);
extern int    LINUXAPI capi_manufacturer(UINT cmd, LPVOID data);

static WORD _capi_isinstalled(void)
{
	return(0);
}

static WORD _capi_register(capi_register_params* parms, LPWORD applid)
{
	return(capi_register(parms, applid));
}

static WORD _capi_release(WORD applid)
{
	return(capi_release(applid));
}

typedef struct {
	WORD   length;
	WORD   applId;
	WORD   command;
	WORD   msgNum;
	DWORD  ncci;
#if defined(CONFIG_X86_64)
	DWORD  data32;
#else
	LPVOID data32;
#endif
	WORD   dataLength;
	WORD   dataHandle;
	WORD   dataFlags;
#if defined(CONFIG_X86_64)
	LPVOID data64;
#endif
	} __attribute((packed)) CAPIMSG;

static WORD _capi_put_message(WORD applid, struct sk_buff* skb)
{
	CAPIMSG* msg = (CAPIMSG*)skb->data;
	CAPIMSG  tmpMsg;
	WORD     res;

	if (msg->command == 0x8086) {
		memcpy(&tmpMsg, msg, sizeof(CAPIMSG));
		msg = &tmpMsg;

#if defined(CONFIG_X86_64)
		msg->data32 = 0;
		msg->data64 = (LPBYTE)skb->data + msg->length;
#else
		msg->data32 = (LPBYTE)skb->data + msg->length;
#endif
		}
	res = capi_put_message_helper(applid, msg);
	kfree_skb(skb);
	return(res);
}

static WORD _capi_get_message(WORD applid, struct sk_buff** skb)
{
	LPVOID msg;
	WORD   l1, l2, err;

	while ((err = capi_get_message_helper(applid, &msg)) == 0) {
		l1 = ((CAPIMSG*)msg)->length;
		l2 = (((CAPIMSG*)msg)->command == 0x8286 ? ((CAPIMSG*)msg)->dataLength : 0);
		if ((*skb = alloc_skb(l1+l2, GFP_ATOMIC)) != NULL) {
			memcpy(skb_put(*skb, l1), msg, l1);
			if (l2 != 0) {
#if defined(CONFIG_X86_64)
				memcpy(skb_put(*skb, l2), ((CAPIMSG*)msg)->data64, l2);
#else
				memcpy(skb_put(*skb, l2), ((CAPIMSG*)msg)->data32, l2);
#endif
				}
			return(0);
			}
		}
	return(err);
}

static SIGNALFUNC signalMap[256];

static void LINUXAPI _signalFunc(WORD applid, LPVOID parm)
{
	signalMap[applid](applid, parm);
}

static WORD _capi_set_signal(WORD applid, SIGNALFUNC signal, LPVOID parm)
{
	signalMap[applid] = signal;
	return(capi_set_signal(applid, _signalFunc, parm));
}

static WORD _capi_get_manufacturer(UINT ctrl, BYTE buf[CAPI_MANUFACTURER_LEN])
{
	return(capi_get_manufacturer(ctrl, buf));
}

static WORD _capi_get_version(UINT ctrl, struct capi_version* ver)
{
	return(capi_get_version(ctrl, ver));
}

static WORD _capi_get_serial(UINT ctrl, BYTE serial[CAPI_SERIAL_LEN])
{
	return(capi_get_serial(ctrl, serial));
}

static WORD _capi_get_profile(UINT ctrl, struct capi_profile* prof)
{
	return(capi_get_profile(ctrl, prof));
}

static int  _capi_manufacturer(UINT cmd, LPVOID data)
{
	return(capi_manufacturer(cmd, data));
}

struct capi_interface {
	__u16 (*capi_isinstalled)(void);
	__u16 (*capi_register)(capi_register_params* rparam, __u16* applidp);
	__u16 (*capi_release)(__u16 applid);
	__u16 (*capi_put_message)(__u16 applid, struct sk_buff* msg);
	__u16 (*capi_get_message)(__u16 applid, struct sk_buff** msgp);
	__u16 (*capi_set_signal)(__u16 applid, void (*signal)(__u16 applid, LPVOID param), LPVOID param);
	__u16 (*capi_get_manufacturer)(__u32 contr, __u8 buf[CAPI_MANUFACTURER_LEN]);
	__u16 (*capi_get_version)(__u32 contr, struct capi_version* verp);
	__u16 (*capi_get_serial)(__u32 contr, __u8 serial[CAPI_SERIAL_LEN]);
	__u16 (*capi_get_profile)(__u32 contr, struct capi_profile* profp);
	int   (*capi_manufacturer)(unsigned int cmd, LPVOID data);
	};

struct capi_interface_user {
	char                        name[20];
	void                        (*callback)(unsigned cmd, __u32 ctrl, LPVOID data);
	struct capi_interface_user* next;
	};

static struct capi_interface capiInterface = {
	_capi_isinstalled,
	_capi_register,
	_capi_release,
	_capi_put_message,
	_capi_get_message,
	_capi_set_signal,
	_capi_get_manufacturer,
	_capi_get_version,
	_capi_get_serial,
	_capi_get_profile,
	_capi_manufacturer,
	};

static struct capi_interface_user* capi_users      = NULL;
static spinlock_t                  capi_users_lock = __SPIN_LOCK_UNLOCKED(capi_users_lock);

struct capi_interface* attach_capi_interface(struct capi_interface_user* user)
{
	struct capi_interface*      res = NULL;
	struct capi_interface_user* u;

	spin_lock(&capi_users_lock);
	for (u = capi_users; u != NULL && u != user; u = u->next)
		;
	if (u == NULL) {
		user->next = capi_users;
		capi_users = user;
		res        = &capiInterface;
		}
	spin_unlock(&capi_users_lock);
	return(res);
}

int detach_capi_interface(struct capi_interface_user* user)
{
	struct capi_interface_user* prev;
	struct capi_interface_user* cur;
	int                         res = -1;

	spin_lock(&capi_users_lock);
	for (prev = NULL, cur = capi_users; cur != NULL; prev = cur, cur = cur->next) {
		if (cur == user) {
			if (prev == NULL)
				capi_users = cur->next;
			else
				prev->next = cur->next;
			cur->next = NULL;
			res       = 0;
			break;
			}
		}
	spin_unlock(&capi_users_lock);
	return(res);
}

void LINUXAPI NotifyInterfaces(unsigned cmd, __u32 ctrl, LPVOID data)
{
	struct capi_interface_user* cur;

	spin_lock(&capi_users_lock);
	for (cur = capi_users; cur != NULL; cur = cur->next) {
		if (cur->callback != NULL)
			(cur->callback)(cmd, ctrl, data);
		}
	spin_unlock(&capi_users_lock);
}

EXPORT_SYMBOL(attach_capi_interface);
EXPORT_SYMBOL(detach_capi_interface);

///////////////////////////////////////////////////////////////////////////////
// FIO - Device File I/O Callbacks

static DWORD linuxInterface[] = {
	CAPI_REGISTER,         sizeof(capi_register_params),  0,
	CAPI_GET_MANUFACTURER, sizeof(DWORD),                 CAPI_MANUFACTURER_LEN,
	CAPI_GET_VERSION,      sizeof(DWORD),                 sizeof(capi_version),
	CAPI_GET_SERIAL,       sizeof(DWORD),                 CAPI_SERIAL_LEN,
	CAPI_GET_PROFILE,      sizeof(DWORD),                 sizeof(capi_profile),
	CAPI_MANUFACTURER_CMD, sizeof(capi_manufacturer_cmd), sizeof(capi_manufacturer_cmd),
	CAPI_GET_ERRCODE,      0,                             sizeof(WORD),
	CAPI_INSTALLED,        0,                             sizeof(WORD),
	CAPI_GET_FLAGS,        0,                             sizeof(DWORD),
	CAPI_SET_FLAGS,        sizeof(DWORD),                 0,
	CAPI_CLR_FLAGS,        sizeof(DWORD),                 0,
	CAPI_NCCI_OPENCOUNT,   sizeof(DWORD),                 0,
	CAPI_NCCI_GETUNIT,     sizeof(DWORD),                 0,
	0};

static ssize_t OS_OnRead(struct file* fp, char __user* buf, size_t size, loff_t* offset)
{
	return(ExecuteIrp(fp, fp->private_data, IRP_MJ_READ, IRP_NO_MINOR, NULL, 0, buf, size));
}

static ssize_t OS_OnWrite(struct file* fp, const char __user* buf, size_t size, loff_t* offset)
{
	int res;

	if ((res = ExecuteIrp(fp, fp->private_data, IRP_MJ_WRITE, IRP_NO_MINOR, buf, size, NULL, 0)) > 0)
     	*offset += res;
	return(res);
}

static int OS_OnOpen(struct inode* node, struct file* fp)
{
	int        res;
	PRIV_DATA* data;

	if ((data = AllocFileData(NULL, MINOR(node->i_rdev))) == NULL) {
		DbgTrace("OS_OnOpen: AllocPrivateDate failed");
		res = -1;
		}
	else if ((res = ExecuteIrp(fp, data, IRP_MJ_CREATE, IRP_NO_MINOR, NULL, 0, NULL, 0)) < 0)
		FreeFileData(data);
	else {
		data->vptr[0]    = fp;
  		fp->private_data = data;
  		}
	return(res);
}

static int OS_OnRelease(struct inode* node, struct file* fp)
{
	int    res;
	LPVOID data;

	if ((res = ExecuteIrp(fp, fp->private_data, IRP_MJ_CLOSE, IRP_NO_MINOR, NULL, 0, NULL, 0)) >= 0) {
		if ((data = fp->private_data) != NULL) {
			fp->private_data = NULL;
			FreeFileData(data);
			}
		}
	return(res);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18)
static int OS_OnFlush(struct file* fp, fl_owner_t id)
#else
static int OS_OnFlush(struct file* fp)
#endif
{
	return(ExecuteIrp(fp, fp->private_data, IRP_MJ_FLUSH, IRP_NO_MINOR, NULL, 0, NULL, 0));
}

DEFINE_MEM_POOL(poolPoll, wait_queue_head_t, 128);

void* LINUXAPI LWaitQueueInit()
{
	wait_queue_head_t* entry;

	if ((entry = (wait_queue_head_t*)LMemPoolAlloc(&poolPoll)) != NULL)
		init_waitqueue_head(entry);
	return(entry);
}

void LINUXAPI LWaitQueueDestroy(void* obj)
{
	// BUGBUGBUG: remove entry from lists???
	LMemPoolFree(&poolPoll, obj);
}

void LINUXAPI LWaitQueueWait(void* obj, Params* irp)
{
	wait_queue_head_t* entry = (wait_queue_head_t*)obj;
	PRIV_DATA*         priv  = (PRIV_DATA*)irp->pFileHdl;
	poll_table*        wait	 = (poll_table*)irp->Context;

	poll_wait(priv->vptr[0], entry, wait);
}

void LINUXAPI LWaitQueueWake(void* obj)
{
	wake_up_interruptible((wait_queue_head_t*)obj);
}

static unsigned int OS_OnPoll(struct file* fp, poll_table* wait)
{
	int     res;
	Params* irp;

	res = -EFAULT;
	if ((irp = LAllocateIrp(fp->private_data, IRP_MJ_POLL, IRP_NO_MINOR, NULL, 0, 0)) != NULL) {
		res = -ERESTARTSYS;
		if (down_interruptible(&userSema) == 0) {
			irp->Context = wait;
			(*IRPHandler)(irp);
			if (irp->status == STATUS_CANCELLED)
				res = -ERESTARTSYS;
			else if (irp->status == ERROR_INVALID_PARAMETER)
				res = -EINVAL;
			else if (irp->status != STATUS_SUCCESS)
				res = -EIO;
			else if (irp->ioCnt == 0)
				res = POLLOUT|POLLWRNORM;
			else
				res = POLLOUT|POLLWRNORM|POLLIN|POLLRDNORM;
			up(&userSema);
			}
		LFreeIrp(irp);
		}
	return(res);
}

typedef struct _IOCTL_DATA {
	void __user* pInBuf;
	void __user* pOutBuf;
	DWORD        nInBufSize;
	DWORD        nOutBufSize;
	DWORD        nRetBytes;
	} __attribute__((packed)) IOCTL_DATA;

#if defined(CONFIG_COMPAT)

typedef struct _IOCTL32_DATA {
	DWORD        pInBuf;
	DWORD        pOutBuf;
	DWORD        nInBufSize;
	DWORD        nOutBufSize;
	DWORD        nRetBytes;
	} __attribute((packed)) IOCTL32_DATA;

static long ioctlHelper(struct file* fp, unsigned int cmd, unsigned long arg, BOOL ioctl32)
{
	IOCTL32_DATA       data32;
	IOCTL_DATA         data;
	int                idx, res;
	DWORD              major;

	for (idx = 0; linuxInterface[idx] != 0 && linuxInterface[idx] != cmd; idx += 3)
		;

	if (linuxInterface[idx] != 0) {
		data.pInBuf      = (void __user*)arg;
		data.pOutBuf     = (void __user*)arg;
		data.nInBufSize  = linuxInterface[idx+1];
		data.nOutBufSize = linuxInterface[idx+2];
		major            = IRP_MJ_INTERNAL_DEVICE_CONTROL;
		}
	else if (ioctl32 && access_ok(VERIFY_WRITE, (void __user*)arg, sizeof(IOCTL32_DATA)) && copy_from_user(&data32, (void __user*)arg, sizeof(IOCTL32_DATA)) == 0) {
		data.pInBuf      = (void __user*)(unsigned long)data32.pInBuf;
		data.pOutBuf     = (void __user*)(unsigned long)data32.pOutBuf;
		data.nInBufSize  = data32.nInBufSize;
		data.nOutBufSize = data32.nOutBufSize;
		major            = IRP_MJ_DEVICE_CONTROL;
		}
	else if (!ioctl32 && access_ok(VERIFY_WRITE, (void __user*)arg, sizeof(IOCTL_DATA)) && copy_from_user(&data, (void __user*)arg, sizeof(IOCTL_DATA)) == 0)
		major = IRP_MJ_DEVICE_CONTROL;
	else
		return(-EFAULT);

	if (data.nInBufSize != 0 && !access_ok(VERIFY_READ, data.pInBuf, data.nInBufSize))
		res = -EFAULT;
	else if (data.nOutBufSize != 0 && !access_ok(VERIFY_WRITE, data.pOutBuf, data.nOutBufSize))
		res = -EFAULT;
	else if ((res = ExecuteIrp(fp, fp->private_data, major, cmd, data.pInBuf, data.nInBufSize, data.pOutBuf, data.nOutBufSize)) >= 0 && linuxInterface[idx] == 0 && put_user(res, (ioctl32 ? &((IOCTL32_DATA __user*)arg)->nRetBytes : &((IOCTL_DATA __user*)arg)->nRetBytes)) != 0)
		res = -EFAULT;

	return(res);
}

#if HAVE_UNLOCKED_IOCTL
static long OS_OnIoctl(struct file* fp, unsigned int cmd, unsigned long arg)
#else
static int OS_OnIoctl(struct inode* node, struct file* fp, unsigned int cmd, unsigned long arg)
#endif
{
	return(ioctlHelper(fp, cmd, arg, FALSE));
}

#if HAVE_COMPAT_IOCTL
static long OS_OnCompatIoctl(struct file* fp, unsigned int cmd, unsigned long arg)
#else
static int OS_OnIoctl32(unsigned int fd, unsigned int cmd, unsigned long arg, struct file* fp)
#endif
{
	return(ioctlHelper(fp, cmd, arg, TRUE));
}

#else

#if HAVE_UNLOCKED_IOCTL
static long OS_OnIoctl(struct file* fp, unsigned int cmd, unsigned long arg)
#else
static int OS_OnIoctl(struct inode* node, struct file* fp, unsigned int cmd, unsigned long arg)
#endif
{
	IOCTL_DATA         data;
	int                idx, res;
	DWORD              major;

	for (idx = 0; linuxInterface[idx] != 0 && linuxInterface[idx] != cmd; idx += 3)
		;

	if (linuxInterface[idx] != 0) {
		data.pInBuf      = (void __user*)arg;
		data.pOutBuf     = (void __user*)arg;
		data.nInBufSize  = linuxInterface[idx+1];
		data.nOutBufSize = linuxInterface[idx+2];
		major            = IRP_MJ_INTERNAL_DEVICE_CONTROL;
		}
	else if (access_ok(VERIFY_WRITE, (void __user*)arg, sizeof(IOCTL_DATA)) && copy_from_user(&data, (void __user*)arg, sizeof(IOCTL_DATA)) == 0)
		major = IRP_MJ_DEVICE_CONTROL;
	else
		return(-EFAULT);

	if (data.nInBufSize != 0 && !access_ok(VERIFY_READ, data.pInBuf, data.nInBufSize))
		res = -EFAULT;
	else if (data.nOutBufSize != 0 && !access_ok(VERIFY_WRITE, data.pOutBuf, data.nOutBufSize))
		res = -EFAULT;
	else if ((res = ExecuteIrp(fp, fp->private_data, major, cmd, data.pInBuf, data.nInBufSize, data.pOutBuf, data.nOutBufSize)) >= 0 && linuxInterface[idx] == 0 && __put_user(res, &((IOCTL_DATA __user*)arg)->nRetBytes) != 0)
		res = -EFAULT;

	return(res);
}

#endif

///////////////////////////////////////////////////////////////////////////////
// FIO - Device file setup stuff

static struct file_operations OS_fops = {
	owner:          THIS_MODULE,
	read:           OS_OnRead,
	write:          OS_OnWrite,
	open:           OS_OnOpen,
	release:        OS_OnRelease,
	flush:          OS_OnFlush,
	poll:           OS_OnPoll,

#if HAVE_UNLOCKED_IOCTL
	unlocked_ioctl: OS_OnIoctl,
#else
	ioctl:          OS_OnIoctl,
#endif

#if defined(CONFIG_COMPAT) && HAVE_COMPAT_IOCTL
	compat_ioctl:   OS_OnCompatIoctl,
#endif

	};

STATUS LINUXAPI LFioRegCharDrv(UINT major, LPCSTR name)
{
#if defined(__KERNEL24__)
   return(devfs_register_chrdev(major, name, &OS_fops) < 0 ? ERROR_REG_DEVDRV : STATUS_SUCCESS);
#else
	return(register_chrdev(major, name, &OS_fops) ? ERROR_REG_DEVDRV : STATUS_SUCCESS);
#endif
}

void LINUXAPI LFioUnRegCharDrv(UINT major, LPCSTR name)
{
#if defined(__KERNEL24__)
   devfs_unregister_chrdev(major, name);
#else
	unregister_chrdev(major, name);
#endif
}

#if defined(CONFIG_COMPAT) && !HAVE_COMPAT_IOCTL

#include <asm/ioctl32.h>

void LINUXAPI LRegisterIoctl32Conversion(DWORD code)
{
	register_ioctl32_conversion(code, OS_OnIoctl32);
}

void LINUXAPI LUnregisterIoctl32Conversion(DWORD code)
{
	unregister_ioctl32_conversion(code);
}

#else

void LINUXAPI LRegisterIoctl32Conversion(DWORD code)
{
}

void LINUXAPI LUnregisterIoctl32Conversion(DWORD code)
{
}

#endif

///////////////////////////////////////////////////////////////////////////////
// FS - Support for Linux FileSystem Stuff

#define NODBG               do {}  while(0)
#define TTYDBG              NODBG // LinuxDbgPrint(6, "%s:%d", __PRETTY_FUNCTION__, __LINE__)
#define FSDBG(x)            NODBG // LinuxDbgPrint x

#define CAPIFS_SUPER_MAGIC (('C'<<8)|'N')
#define CAPIFS_SBI_MAGIC   (('C'<<24)|('A'<<16)|('P'<<8)|'I')

#define FS_NAMELEN         31

typedef struct fsobj {
	struct fsobj* next;
	struct inode* inode;
	LPVOID        desc;
	char          name[FS_NAMELEN+1];
	} fsObject;

typedef struct {
	DWORD                magic;
	struct super_block*  next;
	int                  setuid, setgid;
	unsigned int         last_ino;
	uid_t                uid;
	gid_t                gid;
	umode_t              mode;
	fsObject*            firstObj;
	fsObject*            lastObj;
	} mySuperblock;

static mySuperblock* LINUXAPI mysb_Initialize(void)
{
	mySuperblock* res;

	if ((res = (mySuperblock*)HeapAllocate(sizeof(mySuperblock))) != NULL) {
		res->magic    = CAPIFS_SBI_MAGIC;
		res->next     = NULL;
		res->setuid   = 0;
		res->setgid   = 0;
		res->uid      = 0;
		res->gid      = 0;
		res->mode     = 0666;
		res->firstObj = NULL;
		res->lastObj  = NULL;
		res->last_ino = 0;
		}
	return(res);
}

static void LINUXAPI mysb_Destroy(mySuperblock* msb)
{
	HeapFree(msb);
}

static inline mySuperblock* SBI(struct super_block* sb)
{
#if defined(__KERNEL24__)
	return((mySuperblock*)sb->u.generic_sbp);
#else
	return((mySuperblock*)sb->s_fs_info);
#endif
}

static void LINUXAPI mysb_Insert(mySuperblock* sbi, fsObject* obj)
{
	fsObject* tmp;

	for (tmp = sbi->firstObj; tmp != NULL; tmp = tmp->next) {
		if (tmp == obj)
			return;
		}

	obj->next = NULL;
	if (sbi->firstObj == NULL)
		sbi->firstObj = obj;
	else
		sbi->lastObj->next = obj;
	sbi->lastObj = obj;
}

static fsObject* LINUXAPI mysb_Remove(mySuperblock* sbi, LPVOID desc)
{
	fsObject* prev;
	fsObject* cur;

	for (prev = NULL, cur = sbi->firstObj; cur != NULL; prev = cur, cur = cur->next) {
		if (cur->desc == desc) {
			if (prev == NULL)
				sbi->firstObj = cur->next;
			else
				prev->next = cur->next;
			if (cur->next == NULL)
				sbi->lastObj = prev;
			return(cur);
			}
		}
	return(NULL);
}

static void                PutSuperblock(struct super_block* sb);
static int                 RootReadDir(struct file*, LPVOID, filldir_t);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 11, 0)
static int                 RootIterate(struct file*, struct dir_context*);
#endif

#define FSNAME "capifs"

#if defined(__KERNEL24__)
static struct dentry*      RootLookup(struct inode*, struct dentry*);
static int                 Revalidate(struct dentry*, int);
static int                 StatFs(struct super_block* sb, struct statfs* buf);
static struct super_block* ReadSuperblock(struct super_block* s, LPVOID data, int silent);

static struct file_system_type  fsDescriptor        = {name: FSNAME, read_super: ReadSuperblock, fs_flags: 0, owner: THIS_MODULE};
#else

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 6, 0)
static struct dentry*      RootLookup(struct inode*, struct dentry*, unsigned int);
static int                 Revalidate(struct dentry*, unsigned int);
#else
static struct dentry*      RootLookup(struct inode*, struct dentry*, struct nameidata*);
static int                 Revalidate(struct dentry*, struct nameidata*);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 39)
static struct dentry*      MountSuperblock(struct file_system_type* fs_type, int flags, const char* dev_name, void* data);
static int                 StatFs(struct dentry*, struct kstatfs*);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18)
static int                 ReadSuperblock(struct file_system_type*, int xxx, const char* name, LPVOID data, struct vfsmount*);
static int                 StatFs(struct dentry*, struct kstatfs*);
#else
static struct super_block* ReadSuperblock(struct file_system_type* fs, int xxx, const char* name, LPVOID data);
static int                 StatFs(struct super_block*, struct kstatfs*);
#endif

static struct file_system_type  fsDescriptor = {
	name:     FSNAME,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 39)
	mount:    MountSuperblock,
#else
	get_sb:   ReadSuperblock,
#endif
	kill_sb:  kill_anon_super,
	fs_flags: 0,
	owner:    THIS_MODULE,
	};
#endif

static struct super_block*      mounts              = NULL;
static struct inode_operations  rootInodeOperations = {lookup: RootLookup};
static struct dentry_operations dentryOperations    = {d_revalidate: Revalidate};
static struct super_operations  superOperations     = {put_super: PutSuperblock, statfs: StatFs};
static int                      fsInitialized       = 0;
static struct file_operations   rootOperations      = {
	read: generic_read_dir,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 11, 0)
	iterate: RootIterate,
#else
	readdir: RootReadDir,
#endif
	};

#ifndef GLOBAL_ROOT_UID
#define GLOBAL_ROOT_UID (0)
#define GLOBAL_ROOT_GID (0)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 19, 0)
#define GET_INODE_FROM_FILE(filp)    (filp->f_dentry->d_inode)
#else
#define GET_INODE_FROM_FILE(filp)    (filp->f_inode)
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3, 5, 0))
typedef uid_t kuid_t;
typedef gid_t kgid_t;
#endif

#ifndef CURRENT_TIME
#define CURRENT_TIME current_time(i)
#endif

static struct inode* LINUXAPI my_new_inode(struct super_block* sb, kuid_t uid, kgid_t gid)
{
	mySuperblock* sbi = SBI(sb);
	struct inode* i;

	if ((i = new_inode(sb)) != NULL) {
		i->i_mtime   = CURRENT_TIME;
		i->i_atime   = CURRENT_TIME;
		i->i_ctime   = CURRENT_TIME;
		i->i_blocks  = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
		i->i_blksize = 1024;
#endif
		i->i_uid     = uid;
		i->i_gid     = gid;

		i->i_ino     = ++sbi->last_ino;
		}
	return(i);
}

int LINUXAPI FSInitialize(void)
{
	int err = 0;

	if (fsInitialized++ == 0 && (err = register_filesystem(&fsDescriptor)) != 0)
		fsInitialized = 0;
	FSDBG((6, "%s:  %d/%d", __PRETTY_FUNCTION__, fsInitialized, err));
	return(err);
}

void LINUXAPI FSShutdown(void)
{
	FSDBG((6, "%s: init = %d", __PRETTY_FUNCTION__, fsInitialized));
	if (fsInitialized != 0 && --fsInitialized == 0)
		unregister_filesystem(&fsDescriptor);
	FSDBG((6, "%s: done", __PRETTY_FUNCTION__));
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
	#define CURRENT_UID   current->cred->fsuid
	#define CURRENT_GID   current->cred->fsgid
#else
	#define CURRENT_UID   current->fsuid
	#define CURRENT_GID   current->fsgid
#endif


#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 2, 0)
#define set_nlink(inode, val)          (inode)->i_nlink = (val)
#define drop_nlink(inode)              (inode)->i_nlink--
#endif

void LINUXAPI FSAddObject(LPCSTR name, short major, short minor, LPVOID desc)
{
	fsObject*     obj;
	mySuperblock* sbi;
	struct inode* inode;

	FSDBG((6, "%s: mounts = %p", __PRETTY_FUNCTION__, mounts));
	if (mounts != NULL && (inode = my_new_inode(mounts, GLOBAL_ROOT_UID, GLOBAL_ROOT_GID)) != NULL) {
		sbi = SBI(mounts);

		set_nlink(inode, 1);
		init_special_inode(inode, sbi->mode|S_IFCHR, MKDEV(major, minor));

		obj = (fsObject*)HeapAllocate(sizeof(fsObject));
		strcpy(obj->name, name);
		obj->desc  = desc;
		obj->inode = inode;
		mysb_Insert(sbi, obj);
		}
	FSDBG((6, "%s: done", __PRETTY_FUNCTION__));
}

void LINUXAPI FSRemoveObject(LPVOID desc)
{
	fsObject* obj;

	FSDBG((6, "%s: mounts = %p", __PRETTY_FUNCTION__, mounts));
	if (mounts != NULL && (obj = mysb_Remove(SBI(mounts), desc)) != NULL) {
		drop_nlink(obj->inode);
		iput(obj->inode);
		HeapFree(obj);
		}
	FSDBG((6, "%s: done", __PRETTY_FUNCTION__));
}

static int RootReadDir(struct file* filp, LPVOID dirent, filldir_t filldir)
{
	struct inode* inode = GET_INODE_FROM_FILE(filp);
	mySuperblock* sbi   = SBI(inode->i_sb);
	off_t         nr, idx;
	fsObject*     obj;

	nr = filp->f_pos;
	switch (nr) {
		case 0:
			if (filldir(dirent, ".", 1, nr, inode->i_ino, DT_DIR) < 0)
				break;
			filp->f_pos = ++nr;
			// FALL-THRU

		case 1:
			if (filldir(dirent, "..", 2, nr, inode->i_ino, DT_DIR) < 0)
				break;
			filp->f_pos = ++nr;
			// FALL-THRU

		default:
			for (obj = sbi->firstObj, idx = 2; idx != nr; obj = obj->next, idx++) {
				if (obj == NULL)
					return(0);
				}
			for ( ; obj != NULL; obj = obj->next) {
				if (filldir(dirent, obj->name, strlen(obj->name), nr, obj->inode->i_ino, DT_UNKNOWN) < 0)
					break;
				filp->f_pos = ++nr;
				}
			break;
		}

	return(0);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 11, 0)
static int RootIterate(struct file* filp, struct dir_context* ctx)
{
	int res = RootReadDir(filp, ctx, ctx->actor);
	ctx->pos = filp->f_pos;
	return(res);
}
#endif

#if defined(__KERNEL24__)
static int Revalidate(struct dentry* dentry, int flags)
#elif LINUX_VERSION_CODE < KERNEL_VERSION(3, 6, 0)
static int Revalidate(struct dentry* dentry, struct nameidata* names)
#else
static int Revalidate(struct dentry* dentry, unsigned int xxx)
#endif
{
	mySuperblock* sbi;
	fsObject*     obj;
	off_t         idx;

	if (dentry->d_inode != NULL) {
		sbi = SBI(dentry->d_inode->i_sb);
		for (obj = sbi->firstObj, idx = 2; obj != NULL; obj = obj->next, idx++) {
			if (idx == (off_t)dentry->d_inode->i_ino)
				return(obj->inode == dentry->d_inode);
			}
		}
	return(0);
}

#if defined(__KERNEL24__)
static struct dentry* RootLookup(struct inode* dir, struct dentry* dent)
#elif LINUX_VERSION_CODE < KERNEL_VERSION(3, 6, 0)
static struct dentry* RootLookup(struct inode* dir, struct dentry* dent, struct nameidata* names)
#else
static struct dentry* RootLookup(struct inode* dir, struct dentry* dent, unsigned int xxx)
#endif
{
	mySuperblock* sbi = SBI(dir->i_sb);
	fsObject*     obj;

	dent->d_inode = NULL;
	dent->d_op    = &dentryOperations;
	for (obj = sbi->firstObj; obj != NULL; obj = obj->next) {
		if (strlen(obj->name) == dent->d_name.len && memcmp(obj->name, dent->d_name.name, dent->d_name.len) == 0) {
			if ((dent->d_inode = obj->inode) != NULL)
				atomic_inc(&dent->d_inode->i_count);
			d_add(dent, dent->d_inode);
			break;
			}
		}
	return(NULL);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18)
static int StatFs(struct dentry* entry, struct kstatfs* buf)
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
static int StatFs(struct super_block* sb, struct kstatfs* buf)
#else
static int StatFs(struct super_block* sb, struct statfs* buf)
#endif
{
	buf->f_type    = CAPIFS_SUPER_MAGIC;
	buf->f_bsize   = 1024;
	buf->f_blocks  = 0;
	buf->f_bfree   = 0;
	buf->f_bavail  = 0;
	buf->f_files   = 0;
	buf->f_ffree   = 0;
	buf->f_namelen = FS_NAMELEN;
	return(0);
}

// struct ttyobj;
// struct ttyobj* LINUXAPI TTYOpen(char* name, BYTE major, BYTE minor, LPVOID context);
// void LINUXAPI TTYClose(struct ttyobj*);
// static struct ttyobj* ttyDummy;

static int fillSuperblock(struct super_block* s, LPVOID data, int silent)
{
	mySuperblock*  sbi;
	struct inode*  root_inode;
	struct dentry* root;

	if (s->s_root != NULL)
		return 0;

	if ((sbi = mysb_Initialize()) != NULL) {
#if defined(__KERNEL24__)
		s->u.generic_sbp    = sbi;
#else
		s->s_fs_info        = sbi;
#endif

		s->s_blocksize      = 1024;
		s->s_blocksize_bits = 10;
		s->s_magic          = CAPIFS_SUPER_MAGIC;
		s->s_op             = &superOperations;
		s->s_root           = NULL;

		if ((root_inode = my_new_inode(s, GLOBAL_ROOT_UID, GLOBAL_ROOT_GID)) != NULL) {
			root_inode->i_mode  = S_IFDIR | S_IRUGO | S_IXUGO | S_IWUSR;
			root_inode->i_op    = &rootInodeOperations;
			root_inode->i_fop   = &rootOperations;
			set_nlink(root_inode, 2); // BUGBUGBUG: really?

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 3, 0)
			root = d_make_root(root_inode);
#else
			root = d_alloc_root(root_inode);
#endif

			if (root != NULL) {
				s->s_root = root;
				sbi->next = mounts;
				mounts = s;
				// ttyDummy = TTYOpen("dummy", 191, 1, NULL);
				return 0;
				}

			iput(root_inode);
			}

		mysb_Destroy(sbi);
		}

	return -ENOMEM;
}

#if defined(__KERNEL24__)
static struct super_block* ReadSuperblock(struct super_block* s, LPVOID data, int silent)
{
	return(fillSuperblock(s, data, silent) == 0 ? s : NULL);
}
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 39)
static struct dentry* MountSuperblock(struct file_system_type* fs_type, int flags, const char* dev_name, void* data)
{
	return mount_single(fs_type, flags, data, fillSuperblock);
}
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18)
static int ReadSuperblock(struct file_system_type* fs, int xxx, const char* name, LPVOID data, struct vfsmount* vfs)
{
	return(get_sb_single(fs, xxx, data, fillSuperblock, vfs));
}
#else
static struct super_block* ReadSuperblock(struct file_system_type* fs, int xxx, const char* name, LPVOID data)
{
	return(get_sb_single(fs, xxx, data, fillSuperblock));
}
#endif

static void PutSuperblock(struct super_block* sb)
{
	struct super_block* prev;
	struct super_block* cur;

	// TTYClose(ttyDummy);
	for (prev = NULL, cur = mounts; cur != NULL; prev = cur, cur = SBI(cur)->next) {
		if (cur == sb) {
			if (prev == NULL)
				mounts = SBI(cur)->next;
			else
				SBI(prev)->next = SBI(cur)->next;
			break;
			}
		}
	mysb_Destroy(SBI(sb));
}

///////////////////////////////////////////////////////////////////////////////
// TTY - Support for Linux TTY emulation

#include <linux/tty.h>

#define CAPI_TTY_DRVNAME    "capi_nc"
#define CAPI_TTY_DEVNAME    "capi"
#define CAPI_TTY_MAJOR      191
#define CAPI_TTY_PORTS      256

static struct tty_driver*    capiTTY;
static int                   capiTTYInitialized = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 27)

static struct tty_driver     __capiTTY;

#if defined(__KERNEL24__)
	static int               capiRefCount;
	static struct termios*   capiTermios[CAPI_TTY_PORTS]       = {NULL};
	static struct termios*   capiTermiosLocked[CAPI_TTY_PORTS] = {NULL};
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)
	static struct termios*   capiTermios[CAPI_TTY_PORTS]       = {NULL};
	static struct termios*   capiTermiosLocked[CAPI_TTY_PORTS] = {NULL};
#elif LINUX_VERSION_CODE < KERNEL_VERSION(3, 3, 0)
	static struct ktermios*  capiTermios[CAPI_TTY_PORTS]       = {NULL};
	static struct ktermios*  capiTermiosLocked[CAPI_TTY_PORTS] = {NULL};
#else
	static struct ktermios*  capiTermios[CAPI_TTY_PORTS]       = {NULL};
#endif

static struct tty_struct*    capiTable[CAPI_TTY_PORTS]         = {NULL};

struct tty_driver *alloc_tty_driver(int lines)
{
	memset(&__capiTTY, 0, sizeof(__capiTTY));
	__capiTTY.magic = TTY_DRIVER_MAGIC;
	__capiTTY.num   = lines;
	__capiTTY.owner = THIS_MODULE;

#if defined(__KERNEL24__)
	__capiTTY.refcount       = &capiRefCount;
	__capiTTY.table          = capiTable;
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 28)
	__capiTTY.refcount       = 0;
	__capiTTY.ttys           = capiTable;
#else
	kref_init(&__capiTTY.kref);
	__capiTTY.ttys           = capiTable;
#endif
	__capiTTY.termios        = capiTermios;

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 3, 0)
	__capiTTY.termios_locked = capiTermiosLocked;
#endif

	return(&__capiTTY);
}

#endif

#define MAX_PKTSIZE   2048

typedef struct ttypkt {
	LPBYTE            buf;
	WORD              len;
	} ttyPacket;

struct ttyobj;

#if defined(__KERNEL24__)

struct tty_worker {
	struct tq_struct   kernel_work;
	struct ttyobj*     ttyObj;
	};

typedef void (*LINUXWORKER)(LPVOID ctx);

#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)

struct tty_worker {
	struct work_struct kernel_work;
	struct ttyobj*     ttyObj;
	};

typedef void (*LINUXWORKER)(LPVOID ctx);

#else

struct tty_worker {
	struct work_struct kernel_work;
	struct ttyobj*     ttyObj;
	};

typedef void (*LINUXWORKER)(struct work_struct* ctx);

#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 27)

#define TTY_LDISC_OP(obj, name)       (obj)->tty->ldisc.name

#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 31)

#define TTY_LDISC_OP(obj, name)       (obj)->tty->ldisc.ops->name

#else

#define TTY_LDISC_OP(obj, name)       (obj)->tty->ldisc->ops->name

#endif

typedef struct ttyobj {
	struct ttyobj*     link;
	atomic_t           ttyopencount;
	struct tty_struct* tty;
	struct tty_worker  ttyTask;
	BYTE               minor, scheduled;
	atomic_t           refCount;
	struct timer_list  timer;
	LPVOID             context;
	BOOL               txStopped, rxStopped, txStuffed;
	WORD               avlRead, avlWrite, txRead, txWrite, rxRead, rxWrite, txCount;
	ttyPacket*         txPkt;
	ttyPacket*         avlQueue[8];
	ttyPacket*         txQueue[8];
	ttyPacket*         rxQueue[8];

	BYTE               txBuffer[5*(sizeof(ttyPacket)+MAX_PKTSIZE)];
	} ttyObject;

static ttyObject* objArr[256] = {NULL};
static spinlock_t _tty_lock   = __SPIN_LOCK_UNLOCKED(_tty_lock);
static void*      _tty_owner  = NULL;

extern void LINUXAPI TTYDataB3Req(LPVOID context, ttyPacket* pkt);
extern void LINUXAPI TTYDataB3Resp(LPVOID context, ttyPacket* pkt);
       void LINUXAPI TTYConfirm(ttyObject* obj, ttyPacket* pkt);

static void LINUXAPI ConfirmHelper(ttyObject* obj, ttyPacket* pkt)
{
	obj->txCount                -= pkt->len;
	obj->avlQueue[obj->avlWrite] = pkt;
	obj->avlWrite                = (obj->avlWrite + 1) & 7;
}

static void LINUXAPI unrefObject(ttyObject* obj)
{
	if (atomic_dec_and_test(&obj->refCount)) {
		del_timer(&obj->timer);
		objArr[obj->minor] = NULL;
		obj->minor         = 0;
		FSRemoveObject(obj);
		HeapFree(obj);
		}
}

static ttyObject* LINUXAPI enterProtection(ttyObject* obj)
{
	atomic_inc(&obj->refCount);
	if (_tty_owner != current)
		spin_lock(&_tty_lock);
	return(xchg(&_tty_owner, current));
}

static void LINUXAPI leaveProtection(ttyObject* obj, void* loop)
{
	WORD tmp;

	if (loop == NULL) {
		// Callbacks in den Treiber laufen aus einem Workitem. Dieses Workitem wird nur scheduled, wenn es auch etwas zu tun hat
		tmp = 0;
		if (obj->scheduled == 0 && obj->tty != NULL) {
			tmp  = (obj->txStuffed && obj->avlRead != obj->avlWrite && TTY_LDISC_OP(obj, write_wakeup) != NULL);
			tmp |= (!obj->rxStopped && obj->rxRead != obj->rxWrite);
			tmp |= (obj->context == NULL && obj->txRead != obj->txWrite);
			tmp |= (!obj->txStopped && obj->txRead != obj->txWrite);
			if (tmp) {
				obj->scheduled = 1;
				atomic_inc(&obj->refCount);
				}
			}
		_tty_owner = NULL;
		spin_unlock(&_tty_lock);

		if (tmp) {
#if defined(__KERNEL24__)
			schedule_task((struct tq_struct*)&obj->ttyTask);
#else
			schedule_work((struct work_struct*)&obj->ttyTask);
#endif
			}
		}
	unrefObject(obj);
}

extern void LINUXAPI LEnableFastEvent(void);
extern void LINUXAPI LDisableFastEvent(void);

static void* LINUXAPI EnterProtection(ttyObject* obj)
{
	LDisableFastEvent();
	return(enterProtection(obj));
}

static void LINUXAPI LeaveProtection(ttyObject* obj, void* loop)
{
	leaveProtection(obj, loop);
	LEnableFastEvent();
}

static void LINUXAPI SendPacket(ttyObject* obj)
{
	obj->txQueue[obj->txWrite] = obj->txPkt;
	obj->txWrite               = (obj->txWrite + 1) & 7;
	obj->txPkt                 = NULL;
	del_timer(&obj->timer);
}

static void LINUXAPI SendHelper(ttyObject* obj)
{
	void* loop;

	TTYDBG;
	loop = enterProtection(obj);
	if (obj->txPkt != NULL && obj->txPkt->len != 0)
		SendPacket(obj);
	leaveProtection(obj, loop);
	TTYDBG;
}

static void ttyCallback(struct tty_worker* ctx)
{
	TTYDBG;

	if (down_interruptible(&userSema) == 0) {
		ttyObject* obj = ctx->ttyObj;
		void*      loop;
		ttyPacket* pkt;
		WORD       tmp;

		loop = EnterProtection(obj);
		obj->scheduled = 0;
		unrefObject(obj);

		// Sendedaten verarbeiten
		while (obj->context == NULL && (tmp = obj->txRead) != obj->txWrite) {
			obj->txRead = (obj->txRead + 1) & 7;
			ConfirmHelper(obj, obj->txQueue[tmp]);
			}
		while (!obj->txStopped && (tmp = obj->txRead) != obj->txWrite) {
			obj->txRead = (obj->txRead + 1) & 7;
			TTYDataB3Req(obj->context, obj->txQueue[tmp]);
			}

		// Write Wakeup
		if (obj->txStuffed && obj->avlRead != obj->avlWrite && obj->tty != NULL && TTY_LDISC_OP(obj, write_wakeup) != NULL) {
			TTYDBG;
			obj->txStuffed = 0;
			LeaveProtection(obj, loop);
			TTY_LDISC_OP(obj, write_wakeup)(obj->tty);
			loop = EnterProtection(obj);
			}

		// Empfangene Daten verarbeiten
		while (!obj->rxStopped && obj->rxRead != obj->rxWrite && obj->tty != NULL && TTY_LDISC_OP(obj, receive_buf) != NULL) {
			TTYDBG;
			pkt = obj->rxQueue[obj->rxRead];

	#if defined(__KERNEL24__)
			LeaveProtection(obj, loop);
			if (TTY_LDISC_OP(obj, receive_room) != NULL && TTY_LDISC_OP(obj, receive_room)(obj->tty) < pkt->len) {
				mod_timer(&obj->timer, 4);
				return;
				}
			loop = EnterProtection(obj);
	#endif

			obj->rxRead = (obj->rxRead + 1) & 7;

			LeaveProtection(obj, loop);
			TTY_LDISC_OP(obj, receive_buf)(obj->tty, pkt->buf, 0, pkt->len);
			loop = EnterProtection(obj);
			TTYDataB3Resp(obj->context, pkt);
			}

		// Fertig
		LeaveProtection(obj, loop);

		up(&userSema);
		}

	TTYDBG;
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0)
static void ttyTimer(unsigned long parm)
{
    ttyObject* obj = (ttyObject*)parm;
#else
static void ttyTimer(struct timer_list* t)
{
    ttyObject* obj = from_timer(obj, t, timer);
#endif

	TTYDBG;
	SendHelper(obj);
}

ttyObject* LINUXAPI TTYOpen(char* name, BYTE major, BYTE minor, LPVOID context)
{
	ttyObject* obj;
	ttyPacket* pkt;
	LPBYTE     ptr;

	TTYDBG;

	if ((obj = objArr[minor]) == NULL) {
		obj = (ttyObject*)HeapAllocate(sizeof(ttyObject));
		objArr[minor] = obj;
		FSAddObject(name, major, minor, obj);
		}

	memset(obj, 0, sizeof(ttyObject));

	atomic_set(&obj->refCount, 1);

#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0)
	init_timer(&obj->timer);
	obj->timer.function = ttyTimer;
	obj->timer.data     = obj;
#else
	timer_setup(&obj->timer, ttyTimer, 0);
#endif

	obj->context        = context;
	obj->minor          = minor;

	obj->rxStopped      = TRUE;

	obj->ttyTask.ttyObj = obj;

#if defined(__KERNEL24__)
	obj->ttyTask.kernel_work.routine = (LINUXWORKER)ttyCallback;
	obj->ttyTask.kernel_work.data    = &obj->ttyTask;
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)
	INIT_WORK((struct work_struct*)&obj->ttyTask, (LINUXWORKER)ttyCallback, &obj->ttyTask);
#else
	INIT_WORK((struct work_struct*)&obj->ttyTask, (LINUXWORKER)ttyCallback);
#endif

	for (ptr = obj->txBuffer; ptr < obj->txBuffer+sizeof(obj->txBuffer); ptr += sizeof(ttyPacket)+MAX_PKTSIZE) {
		pkt                            = (ttyPacket*)ptr;
		pkt->buf                       = ptr + sizeof(ttyPacket);
		pkt->len                       = MAX_PKTSIZE;
		obj->avlQueue[obj->avlWrite++] = pkt;
		}

	return(obj);
}

void LINUXAPI TTYClose(ttyObject* obj)
{
	void* loop;

	TTYDBG;

	if (objArr[obj->minor] == obj) {
		loop = enterProtection(obj);
		if (obj->txPkt != NULL && obj->txPkt->len != 0)
			SendPacket(obj);
		while (obj->rxRead != obj->rxWrite) {
			ttyPacket* pkt = obj->rxQueue[obj->rxRead];
			obj->rxRead = (obj->rxRead + 1) & 7;
			TTYDataB3Resp(obj->context, pkt);
			}
		obj->context = NULL;
		if (obj->tty != NULL)
			tty_vhangup(obj->tty);
		unrefObject(obj);
		leaveProtection(obj, loop);
		}
	TTYDBG;
}

void LINUXAPI TTYReceive(ttyObject* obj, ttyPacket* pkt)
{
	WORD  tmp;
	void* loop;

	TTYDBG;

	loop = enterProtection(obj);
	if ((tmp = (obj->rxWrite + 1) & 7) != obj->rxRead) {
		obj->rxQueue[obj->rxWrite] = pkt;
		obj->rxWrite               = tmp;
		}
	else
		TTYDataB3Resp(obj->context, pkt);
	leaveProtection(obj, loop);

	TTYDBG;
}

void LINUXAPI TTYConfirm(ttyObject* obj, ttyPacket* pkt)
{
	void* loop;

	TTYDBG;

	loop = enterProtection(obj);
	ConfirmHelper(obj, pkt);
	leaveProtection(obj, loop);

	TTYDBG;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 28)

static int ctty_install(struct tty_driver* driver, struct tty_struct* tty)
{
	TTYDBG;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)
	if ((tty->port = kmalloc(sizeof(struct tty_port), GFP_KERNEL)) == NULL)
		return -ENOMEM;

	tty_port_init(tty->port);
	// tty_buffer_set_limit(tty->port, 8192);
#endif

	memset(&tty->termios_locked, 0, sizeof(tty->termios_locked));
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0)
	tty->termios = driver->init_termios;
#else
	tty->termios = &driver->init_termios;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 28)
	tty_driver_kref_get(driver);
#endif

	tty->count++;
	driver->ttys[tty->index] = tty;

	return 0;
}

static void ctty_remove(struct tty_driver* driver, struct tty_struct* tty)
{
	TTYDBG;

	driver->ttys[tty->index] = NULL;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 6, 0)
static struct tty_struct* ctty_lookup(struct tty_driver* driver, struct file* filp, int idx)
#else
static struct tty_struct* ctty_lookup(struct tty_driver* driver, struct inode* inode, int idx)
#endif
{
	TTYDBG;

	return driver->ttys[idx];
}

#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 32)

static void ctty_cleanup(struct tty_struct* tty)
{
	TTYDBG;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)
	tty_port_put(tty->port);
#endif
}

#endif

static int ctty_open(struct tty_struct* tty, struct file* filp)
{
	ttyObject* obj;
	void*      loop;

	TTYDBG;

	if ((obj = objArr[MINOR(GET_INODE_FROM_FILE(filp)->i_rdev)]) == NULL) {
		TTYDBG;
		return -ENXIO;
		}

	loop = enterProtection(obj);
	DbgTrace("ctty_open(%p,%d)", obj, obj->ttyopencount);
	tty->driver_data = obj;
	obj->tty         = tty;
	atomic_inc(&obj->ttyopencount);
	atomic_inc(&obj->refCount);
	leaveProtection(obj, loop);
	TTYDBG;
	return(0);
}

static void ctty_close(struct tty_struct* tty, struct file* filp)
{
	ttyObject* obj;
	void*      loop;

	TTYDBG;

	if ((obj = (ttyObject*)tty->driver_data) != NULL) {
		loop = enterProtection(obj);
		DbgTrace("ctty_close(%p,%d)", obj, obj->ttyopencount);
		if (atomic_dec_and_test(&obj->ttyopencount)) {
			tty->driver_data = NULL;
			obj->tty         = NULL;
			}
		unrefObject(obj);
		leaveProtection(obj, loop);
		}
	TTYDBG;
}

static int writeHelper(struct tty_struct* tty, int from_user, const unsigned char* buf, int count, BOOL flush)
{
	ttyObject* obj;
	void*      loop;
	int        res, cnt, total = 0;
	WORD       tmp;

	if ((obj = (ttyObject*)tty->driver_data) != NULL) {
		TTYDBG;
		loop = enterProtection(obj);

		if (flush && obj->txPkt != NULL && obj->txPkt->len != 0)
			SendPacket(obj);

		for ( ; count != 0; count -= cnt, total += cnt) {
			// Falls nötig neues Paket allokatieren
			if (obj->txPkt == NULL) {
				if ((tmp = obj->avlRead) == obj->avlWrite) {
					obj->txStuffed = 1;
					break;
					}
				obj->avlRead    = (tmp + 1) & 7;
				obj->txPkt      = obj->avlQueue[tmp];
				obj->txPkt->len = 0;
				}

			// Maximale Bytezahl für diesen Block ermitteln
			if ((cnt = MAX_PKTSIZE - obj->txPkt->len) > count)
				cnt = count;

			// Daten kopieren
			if (!from_user)
				memcpy(obj->txPkt->buf + obj->txPkt->len, buf, cnt);
			else if ((res = copy_from_user(obj->txPkt->buf + obj->txPkt->len, buf, cnt)) != 0) {
				DbgTrace("ctty_write: copy_from_user failed (%d)", res);
				total = res;
				break;
				}

			// Alles inkrementieren
			obj->txPkt->len += cnt;
			obj->txCount    += cnt;
			buf             += cnt;

			// Block senden, falls er voll ist
			if (flush || obj->txPkt->len >= MAX_PKTSIZE)
				SendPacket(obj);
			}

		if (tty->flags & TTY_DO_WRITE_WAKEUP)
			obj->txStuffed = 1;

		// Timer starten, falls ungesendete Daten da sind
		if (obj->txPkt != NULL && obj->txPkt->len != 0)
			mod_timer(&obj->timer, jiffies + 4);

		leaveProtection(obj, loop);
		}
	TTYDBG;
	return(total);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10)
static int ctty_write(struct tty_struct* tty, int from_user, const unsigned char* buf, int count)
{
	TTYDBG;

	return(writeHelper(tty, from_user, buf, count, TRUE));
}
#else
static int ctty_write(struct tty_struct* tty, const unsigned char* buf, int count)
{
	TTYDBG;

	return(writeHelper(tty, 0, buf, count, TRUE));
}
#endif

static int ctty_write_room(struct tty_struct* tty)
{
	TTYDBG;

	return(255);
}

static void ctty_flush_chars(struct tty_struct* tty)
{
	ttyObject* obj;

	TTYDBG;

	if ((obj = (ttyObject*)tty->driver_data) != NULL)
		SendHelper(obj);
	TTYDBG;
}

static int ctty_chars_in_buffer(struct tty_struct* tty)
{
	ttyObject* obj;
	void*      loop;
	int        res = 0;

	if ((obj = (ttyObject*)tty->driver_data) != NULL) {
		loop = enterProtection(obj);
		res = obj->txCount;
		leaveProtection(obj, loop);
		}
	FSDBG((6, "%s:  %p %d", __PRETTY_FUNCTION__, obj, res));
	return(res);
}

static void LINUXAPI setRxStopped(struct tty_struct* tty, BOOL val)
{
	ttyObject* obj;
	void*      loop;

	if ((obj = (ttyObject*)tty->driver_data) != NULL) {
		loop = enterProtection(obj);
		obj->rxStopped = val;
		leaveProtection(obj, loop);
		}
}

static void LINUXAPI setTxStopped(struct tty_struct* tty, BOOL val)
{
	ttyObject* obj;
	void*      loop;

	if ((obj = (ttyObject*)tty->driver_data) != NULL) {
		loop = enterProtection(obj);
		obj->txStopped = val;
		leaveProtection(obj, loop);
		}
}

#ifndef PPPIOCGCHAN
#define PPPIOCGCHAN 0x80047437
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 39)
static int ctty_ioctl(struct tty_struct* tty, struct file* filp, unsigned int cmd, unsigned long arg)
#else
static int ctty_ioctl(struct tty_struct* tty, unsigned int cmd, unsigned long arg)
#endif
{
	FSDBG((6, "%s: %p %D/%p", __PRETTY_FUNCTION__, tty->driver_data, cmd, arg));

	if (cmd == PPPIOCGCHAN)
		setRxStopped(tty, FALSE);

	return(-ENOIOCTLCMD);
}

static void ctty_throttle(struct tty_struct* tty)
{
	TTYDBG;
	setRxStopped(tty, TRUE);
	TTYDBG;
}

static void ctty_unthrottle(struct tty_struct* tty)
{
	TTYDBG;
	setRxStopped(tty, FALSE);
	TTYDBG;
}

static void ctty_stop(struct tty_struct* tty)
{
	TTYDBG;
	setTxStopped(tty, TRUE);
	TTYDBG;
}

static void ctty_start(struct tty_struct* tty)
{
	TTYDBG;
	setTxStopped(tty, FALSE);
	TTYDBG;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)

static struct tty_operations capiOps = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 28)
	.install         = ctty_install,
	.remove          = ctty_remove,
	.lookup          = ctty_lookup,
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 32)
	.cleanup         = ctty_cleanup,
#endif

	.open            = ctty_open,
	.close           = ctty_close,
	.write           = ctty_write,
	.write_room      = ctty_write_room,
	.flush_chars     = ctty_flush_chars,
	.chars_in_buffer = ctty_chars_in_buffer,
	.ioctl           = ctty_ioctl,
	.throttle        = ctty_throttle,
	.unthrottle      = ctty_unthrottle,
	.stop            = ctty_stop,
	.start           = ctty_start,
	};

#endif

void LINUXAPI TTYInitialize(void)
{
	TTYDBG;

	if (capiTTYInitialized == 0) {
		capiTTYInitialized = 1;

		capiTTY                        = alloc_tty_driver(CAPI_TTY_PORTS);
		capiTTY->type                  = TTY_DRIVER_TYPE_SERIAL;
		capiTTY->subtype               = SERIAL_TYPE_NORMAL;
		capiTTY->init_termios          = tty_std_termios;
		capiTTY->init_termios.c_iflag  = ICRNL;
		capiTTY->init_termios.c_oflag  = OPOST | ONLCR;
		capiTTY->init_termios.c_cflag  = B9600 | CS8 | CREAD | HUPCL | CLOCAL;
		capiTTY->init_termios.c_lflag  = 0;
		capiTTY->init_termios.c_ispeed = tty_termios_input_baud_rate(&capiTTY->init_termios);
		capiTTY->init_termios.c_ospeed = tty_termios_baud_rate(&capiTTY->init_termios);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
		capiTTY->flags                = TTY_DRIVER_REAL_RAW|TTY_DRIVER_RESET_TERMIOS|TTY_DRIVER_NO_DEVFS;
		capiTTY->name                 = CAPI_TTY_DEVNAME "/%d";
#elif LINUX_VERSION_CODE < KERNEL_VERSION(3, 3, 0)
		capiTTY->flags                = TTY_DRIVER_REAL_RAW|TTY_DRIVER_RESET_TERMIOS|TTY_DRIVER_DYNAMIC_DEV;
		capiTTY->name                 = CAPI_TTY_DEVNAME "/";
#else
		capiTTY->flags                = TTY_DRIVER_REAL_RAW|TTY_DRIVER_RESET_TERMIOS|TTY_DRIVER_DYNAMIC_DEV|TTY_DRIVER_DYNAMIC_ALLOC;
		capiTTY->name                 = CAPI_TTY_DEVNAME "/";
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 26)
		capiTTY->ops                  = &capiOps;
#else
		capiTTY->open                 = ctty_open;
		capiTTY->close                = ctty_close;
		capiTTY->write                = ctty_write;
		capiTTY->write_room           = ctty_write_room;
		capiTTY->flush_chars          = ctty_flush_chars;
		capiTTY->chars_in_buffer      = ctty_chars_in_buffer;
		capiTTY->ioctl                = ctty_ioctl;
		capiTTY->throttle             = ctty_throttle;
		capiTTY->unthrottle           = ctty_unthrottle;
		capiTTY->stop                 = ctty_stop;
		capiTTY->start                = ctty_start;
#endif

		capiTTY->driver_name          = CAPI_TTY_DRVNAME;
		capiTTY->major                = CAPI_TTY_MAJOR;
		capiTTY->minor_start          = 0;

		tty_register_driver(capiTTY);
		}
}

void LINUXAPI TTYShutdown(void)
{
	TTYDBG;

	if (capiTTYInitialized != 0) {
		capiTTYInitialized = 0;
		tty_unregister_driver(capiTTY);
		}
}

///////////////////////////////////////////////////////////////////////////////
// INIT - Driver Init and Exit
#if defined(__KERNEL24__)
#define flush_tasks()      flush_scheduled_tasks()
MODULE_PARM(heapFlags, "i");
MODULE_PARM(verbose, "i");
MODULE_PARM(tz, "s");
#else
#define flush_tasks()      flush_scheduled_work()
module_param(heapFlags, int, S_IRUGO);
module_param(verbose, int, S_IRUGO);
module_param(tz, charp, S_IRUGO);
#endif

static void Cleanup(void)
{
	drvActive = FALSE;
	flush_tasks();
	DriverUnload();
	flush_tasks();
}

int OS_DriverOnLoad(void)
{
	DbgTrace("Loading %s", __drvName);

	sema_init(&userSema, 1);
	drvActive = TRUE;
	freeQueue = NULL;
	if (DriverEntry(NULL) < 0) {
		LDbgPrint(KERN_ERR, "error! init library failed!");
		return(-1);
		}
	parseTimezone(&my_tz, tz);
	return(0);
}

void OS_DriverOnUnload(void)
{
//	DbgTrace("OS_DriverOnUnload()");
	Cleanup();
	DbgTrace("Unloaded %s", __drvName);
}

module_init(OS_DriverOnLoad);
module_exit(OS_DriverOnUnload);
